# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom,
#                          and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

def graphviz_output(G, filename, string_function=str):
   """**graphviz_output** (*graph*, *filename*, *string_function* = str)

Writes a graph in the format used by the ``dot`` tool in the graphviz_ package for
graph visualisation.

.. _graphviz: http://www.graphviz.org/

*graph*
   A Gamera ``Graph`` object.

*filename*
   Filename to output to.

*string_function*
   A function to convert node values to strings.  By default this will use Python's
   standard ``str`` function.
"""
   fd = open(filename, 'w')
   if G.is_directed():
      fd.write("digraph G {\n")
      for node in G.get_nodes():
         fd.write('   "%s";\n' % 
                  string_function(node()))
         for edge in node.edges:
            fd.write('   "%s" -> "%s"' % (
               string_function(node()), string_function(edge.to_node())))
            if edge.label is not None:
               fd.write(' [ label = %f ]' % (edge.label))
            fd.write(';\n')
      fd.write("}\n")
   else:
      fd.write("graph G {\n")
      for node in G.get_nodes():
         fd.write('   "%s";\n' % 
                  string_function(node()))
      for edge in G.get_edges():
         fd.write('   "%s" -- "%s"' % (
            string_function(edge.from_node()), string_function(edge.to_node())))
         if edge.label is not None:
            fd.write(' [ label = %f ]' % (edge.label))
         fd.write(";\n")   
      fd.write("}\n")
   fd.close()



#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom,
#                         and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

# This stuff must come the very first before any other gui-specific things
# are imported.

from distutils.version import LooseVersion

# wxPython
try:
   import wxversion
   wxversion.select(["3.0", "2.9", "2.8", "2.6", "2.5", "2.4"])
except ImportError:
   from wxPython.wx import wxVERSION
   # Check that the version is correct
   if wxVERSION[:2] < (2, 4) or wxVERSION[:2] > (3, 0):
     raise RuntimeError("""This version of Gamera requires wxPython 2.4.x, 2.6.x, 2.8.x, 2.9.x, or 3.0.x.  However, it seems that you have wxPython %s installed."""
                      % ".".join([str(x) for x in wxVERSION]))

try:
   from wx import aui
except ImportError:
   aui = None
   
import inspect
import sys, StringIO

from gamera.core import *
from gamera.config import config
   
from gamera import paths, util
from gamera.gui import gamera_display, image_menu, \
     icon_display, classifier_display, var_name, gui_util, \
     image_browser, has_gui

# wxPython
import wx
import wx.py

# Python standard library
# import interactive
import sys, traceback, os, string, os.path, imp

# Set default options
config.add_option(
   "", "--shell-font-face", default=wx.py.editwindow.FACES['mono'],
   help="[shell] Font face used in the shell")
config.add_option(
   "", "--shell-font-size", default=wx.py.editwindow.FACES['size'],
   type="int",
   help="[shell] Font size used in the shell")
config.add_option(
   "-e", "--execfile", type="string", action="append",
   help="[shell] Run execfile(...) on the given file.  This argument may be given multiple times")
config.add_option(
   "-v", "--verbosity-level", type="int", default=1,
   help="[shell] when non zero, debug messages on loaded plugins etc. are written to stdout.")


main_win = None
app = None

######################################################################

class GameraGui:
   def GetImageFilename():
      filename = gui_util.open_file_dialog(None, "*.*")
      if filename:
         return filename
   GetImageFilename = staticmethod(GetImageFilename)

   def ShowImage(image, title, view_function=None, owner=None):
      wx.BeginBusyCursor()
      try:
         img = gamera_display.ImageFrame(title=title, owner=owner)
         img.set_image(image, view_function)
         img.Show(True)
      finally:
         wx.EndBusyCursor()
      return img
   ShowImage = staticmethod(ShowImage)

   def ShowImages(list, view_function=None):
      wx.BeginBusyCursor()
      try:
         img = gamera_display.MultiImageFrame(title = "Multiple Images")
         img.set_image(list, view_function)
         img.Show(1)
      finally:
         wx.EndBusyCursor()
      return img
   ShowImages = staticmethod(ShowImages)

   def ShowHistogram(hist, mark=None):
      f = gamera_display.HistogramDisplay(hist, mark=mark)
      f.Show(1)
   ShowHistogram = staticmethod(ShowHistogram)

   def ShowProjections(x_data, y_data, image):
      f = gamera_display.ProjectionsDisplay(x_data, y_data, image)
      f.Show(1)
   ShowProjections = staticmethod(ShowProjections)

   def ShowClassifier(classifier=None, current_database=[],
                      image=None, symbol_table=[]):
      if classifier is None:
         from gamera import knn
         classifier = knn.kNNInteractive()
      wx.BeginBusyCursor()
      try:
         class_disp = classifier_display.ClassifierFrame(classifier, symbol_table)
         class_disp.set_image(current_database, image)
         class_disp.Show(1)
      finally:
         wx.EndBusyCursor()
      return class_disp
   ShowClassifier = staticmethod(ShowClassifier)

   def UpdateIcons():
      main_win.icon_display.update_icons()
   UpdateIcons = staticmethod(UpdateIcons)

   def TopLevel():
      return main_win
   TopLevel = staticmethod(TopLevel)

   def ProgressBox(message, length=1, numsteps=0):
      return gui_util.ProgressBox(message, length, numsteps)
   ProgressBox = staticmethod(ProgressBox)

if wx.VERSION >= (2, 5):
   import wx.html
   class Calltip(wx.html.HtmlWindow):
      def __init__(self, parent=None, id=-1):
         wx.html.HtmlWindow.__init__(self, parent, id)
         wx.py.crust.dispatcher.connect(receiver=self.display, signal='Shell.calltip')
         if wx.VERSION >= (2, 5) and "gtk2" in wx.PlatformInfo:
            self.SetStandardFonts()
         self.SetBackgroundColour(wx.Colour(255, 255, 232))
         self.message_displayed = False
         self.cache = {}

      def display(self, calltip):
         """Receiver for Shell.calltip signal."""
         html = gui_util.docstring_to_html(calltip)
         self.SetPage(html)
         self.SetBackgroundColour(wx.Colour(255, 255, 232))

      def OnLinkClicked(self, link):
         if not self.message_displayed:
            gui_util.message("Clicking on links is not supported.")
            self.message_displayed = True
else:
   Calltip = wx.py.crust.Calltip

class PyShellGameraShell(wx.py.shell.Shell):
   def __init__(self, *args, **kwargs):
      self.update = None
      wx.py.shell.Shell.__init__(self, *args, **kwargs)
      self.push("from gamera.gui import gui")
      self.push("from gamera.gui.matplotlib_support import *")
      self.push("from gamera.core import *")
      self.push("init_gamera()")
      self.locals = self.interp.locals
      self.autoCallTip = False

      style = wx.py.editwindow.FACES.copy()
      style['mono'] = config.get("shell_font_face")
      style['size'] = config.get("shell_font_size")
      self.setStyles(style)
      self.ScrollToLine(1)
      if wx.VERSION < (2, 5):
         self.autoComplete = False

   def addHistory(self, command):
      if self.update:
         self.update()
      wx.py.shell.Shell.addHistory(self, command)

   def push(self, source):
      wx.py.shell.Shell.push(self, source)
      if source.strip().startswith("import "):
         new_modules = [x.strip() for x in source.strip()[7:].split(",")]
         for module in new_modules:
            if self.interp.locals.has_key(module):
               for obj in self.interp.locals[module].__dict__.values():
                  if (inspect.isclass(obj)):
                     if hasattr(obj, "is_custom_menu"):
                        self.main_win.add_custom_menu(module, obj)
                     elif hasattr(obj, "is_custom_icon_description"):
                        self.main_win.add_custom_icon_description(obj)
         self.update()

   def OnKeyDown(self, event):
      key = event.GetKeyCode()

      if key in (wx.WXK_UP, wx.WXK_DOWN):
         if LooseVersion(wx.__version__) < LooseVersion('3.0'):
             event.m_controlDown = True
         else:
             event.SetControlDown(True)
      wx.py.shell.Shell.OnKeyDown(self, event)

   def GetLocals(self):
      return self.locals

if not aui:
   class PyCrustGameraShell(wx.py.crust.Crust):
      def __init__(self, parent, id=-1, pos=wx.DefaultPosition, 
                   size=wx.DefaultSize, style=0,
                   name='Crust Window', rootObject=None, rootLabel=None,
                   rootIsNamespace=True, intro='', locals=None, 
                   InterpClass=None, *args, **kwds):
         wx.SplitterWindow.__init__(self, parent, id, pos, size, style, name)
         self.shell = PyShellGameraShell(parent=self, introText=intro, 
                                         locals=locals, InterpClass=InterpClass, 
                                         *args, **kwds)
         self.editor = self.shell
         if rootObject is None:
            rootObject = self.shell.interp.locals
         self.notebook = wx.Notebook(parent=self, id=-1, style=wx.NB_BOTTOM)
         self.shell.interp.locals['notebook'] = self.notebook
         self.filling = wx.py.filling.Filling(parent=self.notebook, 
                                           rootObject=rootObject, 
                                           rootLabel=rootLabel, 
                                           rootIsNamespace=rootIsNamespace)
         # Add 'filling' to the interpreter's locals.
         self.shell.interp.locals['filling'] = self.filling
         self.calltip = Calltip(parent=self.notebook)
         self.notebook.AddPage(page=self.calltip, text='Documentation', select=True)
         self.notebook.AddPage(page=self.filling, text='Namespace')
         self.sessionlisting = wx.py.crust.SessionListing(parent=self.notebook)
         self.notebook.AddPage(page=self.sessionlisting, text='History')
         self.SplitHorizontally(self.shell, self.notebook, parent.GetClientSize()[1] - 200)
         self.SetMinimumPaneSize(1)
      
class ShellFrame(wx.Frame):
   def __init__(self, parent, id, title):
      global shell
      wx.Frame.__init__(
         self, parent, id, title, (100, 100),
         # Win32 change
         [600, 550],
         style=wx.DEFAULT_FRAME_STYLE|wx.CLIP_CHILDREN|wx.NO_FULL_REPAINT_ON_RESIZE)
      wx.EVT_CLOSE(self, self._OnCloseWindow)

      self.known_modules = {}
      self.menu = self.make_menu()
      self.SetMenuBar(self.menu)

      if aui:
         self._aui = aui.AuiManager(self)
         nb_style = (aui.AUI_NB_TAB_SPLIT|aui.AUI_NB_TAB_MOVE|
                     aui.AUI_NB_TAB_EXTERNAL_MOVE|aui.AUI_NB_SCROLL_BUTTONS)
         nb = aui.AuiNotebook(self, style=nb_style)
         control_parent = self
      else:
         splitter = wx.SplitterWindow(
            self, -1,
            style=wx.SP_3DSASH|wx.CLIP_CHILDREN|
            wx.NO_FULL_REPAINT_ON_RESIZE|wx.SP_LIVE_UPDATE)
         control_parent = splitter

      self.icon_display = icon_display.IconDisplay(control_parent, self)
      if aui:
         self.shell = PyShellGameraShell(parent=self)
         rootObject = self.shell.interp.locals
         self.filling = wx.py.filling.Filling(parent=nb, 
                                              rootObject=rootObject, 
                                              rootIsNamespace=True)
         # Add 'filling' to the interpreter's locals.
         self.calltip = Calltip(parent=nb)
         self.sessionlisting = wx.py.crust.SessionListing(parent=nb)
      else:
         crust = PyCrustGameraShell(control_parent, -1)
         self.shell = crust.shell
      self.shell.main_win = self
      self.shell.update = self.Update
      image_menu.set_shell(self.shell)
      image_menu.set_shell_frame(self)
      self.shell.push("from gamera.gui import gui")
      self.shell.push("from gamera.gui.matplotlib_support import *")
      self.shell.push("from gamera.core import *")
      self.shell.push("init_gamera()")

      self.shell.update = self.Update
      self.icon_display.shell = self.shell
      self.icon_display.main = self
      self.Update()
      self.shell.SetFocus()

      if aui:
         self._aui.AddPane(
            self.icon_display,
            aui.AuiPaneInfo()
            .Name("icon_display")
            .CenterPane()
            .MinSize(wx.Size(96, 96))
            .CloseButton(False)
            .CaptionVisible(True)
            .Caption("Objects")
            .Dockable(True))
         self._aui.AddPane(
            self.shell,
            aui.AuiPaneInfo()
            .Name("shell")
            .CenterPane()
            .MinSize(wx.Size(200,200))
            .CloseButton(False)
            .CaptionVisible(True)
            .MaximizeButton(True)
            .Caption("Interactive Python Shell")
            .Dockable(True))
         nb.AddPage(self.calltip, "Documentation")
         nb.AddPage(self.filling, "Namespace")
         nb.AddPage(self.sessionlisting, "History")
         self._aui.AddPane(
            nb,
            aui.AuiPaneInfo()
            .Name("notebook")
            .CenterPane()
            .MinSize(wx.Size(200, 100))
            .MaximizeButton(True))
         
         self._aui.GetPane("notebook").Show().Bottom()
         self._aui.GetPane("icon_display").Show().Left()
         self._aui.Update()
      else:
         splitter.SetMinimumPaneSize(20)
         splitter.SplitVertically(self.icon_display, crust, 120)
         splitter.SetSashPosition(120)

      self.status = StatusBar(self)
      self.SetStatusBar(self.status)
      from gamera.gui import gamera_icons
      icon = wx.IconFromBitmap(gamera_icons.getIconBitmap())
      self.SetIcon(icon)
      self.Move(wx.Point(int(30), int(30)))
      wx.Yield()

   def import_command_line_modules(self):
      sys.argv = config.get_free_args()
      if len(sys.argv):
         file = sys.argv[0]
         try:
            name = os.path.basename(file)[:-3]
            module = imp.load_source(name, file)
            self.shell.locals[name] = module
            self.shell.push(name)
            imported = True
         except Exception, e:
            exc_type, exc_value, exc_traceback = sys.exc_info()
            gui_util.message("Error importing file '%s':\n%s" %
                             (file, "".join(
               traceback.format_exception(exc_type, exc_value, exc_traceback))))

      execfiles = config.get("execfile")
      if execfiles is not None:
         for file in execfiles:
            try:
               self.shell.run("execfile(%s)" % repr(file))
            except Exception, e:
               exc_type, exc_value, exc_traceback = sys.exc_info()
               gui_util.message("Error importing file '%s':\n%s" %
                                (file, "".join(traceback.format_exception(
                  exc_type, exc_value, exc_traceback))))

   def make_menu(self):
      self.custom_menus = {}
      file_menu = gui_util.build_menu(
         self,
         (("&Open image...", self._OnFileOpen),
          ("&Image browser...", self._OnImageBrowser),
          (None, None),
          ("Open &XML...", self._OnLoadXML),
          (None, None),
          ("Execute &code...", self._OnExecFile),
          ("&Save history...", self._OnSaveHistory),
          (None, None),
          ("&Biollante...", self._OnBiollante),
          (None, None),
          ("E&xit...", self._OnCloseWindow)))
      classify_menu = gui_util.build_menu(
         self,
         (("&Interactive classifier", self._OnClassifier),))
      toolkits = paths.get_toolkit_names(paths.toolkits)
      self.import_toolkits = {}
      self.reload_toolkits = {}
      self.toolkit_menus = {}
      toolkits_menu = wx.Menu()
      if len(toolkits):
         for toolkit in toolkits:
            toolkitID = wx.NewId()
            toolkit_menu = wx.Menu() #style=wxMENU_TEAROFF)
            toolkit_menu.Append(toolkitID, "Import '%s' toolkit" % toolkit,
                                "Import %s toolkit" % toolkit)
            wx.EVT_MENU(self, toolkitID, self._OnImportToolkit)
            self.import_toolkits[toolkitID] = toolkit
            toolkitID = wx.NewId()
            toolkit_menu.Append(toolkitID, "Reload '%s' toolkit" % toolkit,
                                "Reload %s toolkit" % toolkit)
            wx.EVT_MENU(self, toolkitID, self._OnReloadToolkit)
            self.reload_toolkits[toolkitID] = toolkit
            toolkits_menu.AppendMenu(wx.NewId(), toolkit, toolkit_menu)
            self.toolkit_menus[toolkit] = toolkit_menu
      else:
         toolkits_menu.Append(wx.NewId(), "No toolkits installed.")
      menubar = wx.MenuBar()
      menubar.Append(file_menu, "&File")
      menubar.Append(classify_menu, "&Classify")
      menubar.Append(toolkits_menu, "&Toolkits")
      return menubar

   def add_custom_menu(self, name, menu):
      if name not in self.custom_menus:
         self.toolkit_menus[name].AppendSeparator()
         menu(self.toolkit_menus[name], self.shell, self.shell.GetLocals())
         self.custom_menus.append(name)

   def add_custom_icon_description(self, icon_description):
      self.icon_display.add_class(icon_description)

   def _OnFileOpen(self, event):
      filename = gui_util.open_file_dialog(
         self, util.get_file_extensions("load"))
      if filename:
         name = var_name.get("image", self.shell.locals)
         if name:
            try:
               wx.BeginBusyCursor()
               self.shell.run('%s = load_image(r"%s")' % (name, filename))
            finally:
               wx.EndBusyCursor()

   def _OnImageBrowser(self, event):
      browser = image_browser.ImageBrowserFrame()
      browser.Show(1)

   def _OnLoadXML(self, event):
      from gamera import gamera_xml
      filename = gui_util.open_file_dialog(self, gamera_xml.extensions)
      if filename:
         name = var_name.get("glyphs", self.shell.locals)
         if name:
            wx.BeginBusyCursor()
            try:
               self.shell.run("from gamera import gamera_xml")
               self.shell.run('%s = gamera_xml.glyphs_from_xml(r"%s")' %
                              (name, filename))
            finally:
               wx.EndBusyCursor()

   def _OnExecFile(self, event):
      filename = gui_util.open_file_dialog(self, "Python files (*py)|*.py")
      if filename:
         self.shell.run("execfile(%s)" % (repr(filename)))

   def _OnSaveHistory(self, event):
      filename = gui_util.save_file_dialog(self, "Python files (*.py)|*.py")
      if filename:
         fd = open(filename, "w")
         fd.write(self.history.GetText())
         fd.close()

   def _OnBiollante(self, event):
      from gamera.gui.gaoptimizer import OptimizerFrame
      frame = OptimizerFrame.OptimizerFrame(None, -1, "Biollante - GA Optimization for kNN")
      frame.Show(True)

   def _OnClassifier(self, event):
      name = var_name.get("classifier", self.shell.locals)
      if name:
         self.shell.run("from gamera import knn")
         self.shell.run("%s = knn.kNNInteractive()" % name)
         self.shell.run("%s.display()" % name)

   def _OnImportToolkit(self, event):
      self.shell.run("from gamera.toolkits import %s\n" %
                     self.import_toolkits[event.GetId()])

   def _OnReloadToolkit(self, event):
      self.shell.run("reload(%s)\n" %
                     self.reload_toolkits[event.GetId()])

   def _OnCloseWindow(self, event):
      for window in self.GetChildren():
         if isinstance(window, wx.Frame):
            if not window.Close():
               return
      self.Destroy()
      #app.ExitMainLoop()
      wx.Exit()

   def Update(self):
      self.icon_display.update_icons(self.shell.interp.locals)

class CustomMenu:
   is_custom_menu = 1
   _items = []
   def __init__(self):
      if not has_gui.has_gui:
         return
      main_win = has_gui.gui.TopLevel()
      if not main_win:
         return
      name = self.__class__.__module__.split('.')[-1]
      if not main_win.custom_menus.has_key(name):
         main_win.custom_menus[name] = None
         self.shell = main_win.shell
         self.locals = main_win.shell.locals
         menu = main_win.toolkit_menus[name]
         menu.AppendSeparator()
         if self._items == []:
            menu.Append(wx.NewId(), "--- empty ---")
         else:
            for item in self._items:
               if item == "-":
                  menu.Break()
               else:
                  menuID = wx.NewId()
                  menu.Append(menuID, item)
                  wx.EVT_MENU(main_win, menuID,
                           getattr(self, "_On" +
                                   util.string2identifier(item)))

class StatusBar(wx.StatusBar):
   def __init__(self, parent):
      wx.StatusBar.__init__(self, parent, -1)
      self.SetFieldsCount(3)
      self.SetStatusText("Gamera", 0)

class GameraSplash(wx.SplashScreen):
   def __init__(self):
      from gamera.gui import gamera_icons
      wx.SplashScreen.__init__(self, gamera_icons.getGameraSplashBitmap(),
                              wx.SPLASH_CENTRE_ON_SCREEN|wx.SPLASH_NO_TIMEOUT,
                              1000, None, -1,
                              style = (wx.SIMPLE_BORDER|
                                       wx.FRAME_NO_TASKBAR|wx.STAY_ON_TOP))

def _show_shell():
   global main_win
   main_win = ShellFrame(None, -1, "Gamera")
   main_win.Show(True)
   main_win.import_command_line_modules()

def run(startup=_show_shell):
   global app
   has_gui.has_gui = has_gui.WX_GUI
   has_gui.gui = GameraGui
   from gamera.gui import args_gui

   class MyApp(wx.App):
      def __init__(self, parent):
         #self._startup = startup
         wx.App.__init__(self, 0)
         self.SetExitOnFrameDelete(1)

      # wxWindows calls this method to initialize the application
      def OnInit(self):
         self.SetAssertMode(wx.PYAPP_ASSERT_SUPPRESS)
         self.SetAppName("Gamera")
         self.splash = GameraSplash()
         self.splash.Show()
         wx.Yield()
         init_gamera()
         self.splash.Show(0)
         del self.splash
         return True

      def OnExit(self):
         pass

   app = MyApp(0)
   startup() # must be called here and not in MyApp.OnInit to avoid errors on MacOS X
   app.MainLoop()

if __name__ == "__main__":
   run()




from gamera.core import *
init_gamera()
from gamera import gendoc
from gamera import plugin

# This is a fairly complex test harness, which generates
# a "wrapper" call around every plugin so each plugin can
# be tested separately by py.test

class PluginTester(gendoc.PluginDocumentationGenerator):
   def __init__(self):
      self.images = self.get_generic_images()
      self.methods = plugin._methods_flatten(plugin.plugin_methods)

class TestPlugins:
   pass

def make_test(tester, method):
   def test(self):
      tester.run_example(method, tester.images)
   return test

tester = PluginTester()
for name, method in tester.methods:
   setattr(TestPlugins, "test_plugin_" + name, make_test(tester, method))



import py.test

from gamera.core import *
init_gamera()

#
# Tests for Voronoi tesselation and Delaunay triangulation
#

from gamera.plugins.geometry import delaunay_from_points

#
# input parameter check
#
def test_wrongparams():
    def _voronoi_from_points_input(points,labels):
        image = Image((0,0),(20,20))
        image.voronoi_from_points(points,labels)
    # number of points and labels must match
    py.test.raises(Exception, _voronoi_from_points_input, [(2,2),(5,5)], [2,3,4])
    # labels must be int's and points must be coordinates
    py.test.raises(Exception, _voronoi_from_points_input, [(2,2),(5,5)], [(2,2),(5,5)])
    # TODO: no plausi check in Gamera's passing of PointVector => crash
    #py.test.raises(Exception, _voronoi_from_points_input, [2,,5], [2,5])

#
# voronoi tesselation by image labeling
#
def test_voronoi_cell_labeling():
    img = Image((0,0),(9,9))
    points = [(2,2),(3,2),(3,8),(6,5)]
    labels = [2,2,3,4]
    for i in range(len(points)):
        img.set(points[i],labels[i])
    # voronoi_from_labeled_image
    voronoi1 = img.voronoi_from_labeled_image()
    assert 2 == voronoi1.get((2,4))
    assert 3 == voronoi1.get((4,8))
    assert 4 == voronoi1.get((5,4))
    labelpairs = voronoi1.labeled_region_neighbors()
    assert [2,4] in labelpairs or [4,2] in labelpairs
    assert [2,3] in labelpairs or [3,2] in labelpairs
    assert [4,3] in labelpairs or [3,4] in labelpairs
    # voronoi_from_points
    voronoi2 = Image(img)
    voronoi2.voronoi_from_points(points,labels)
    assert 2 == voronoi2.get((2,4))
    assert 3 == voronoi2.get((4,8))
    assert 4 == voronoi2.get((5,4))
    assert [2,4] in labelpairs or [4,2] in labelpairs
    assert [2,3] in labelpairs or [3,2] in labelpairs
    assert [4,3] in labelpairs or [3,4] in labelpairs

#
# delaunay triangulation
#
def test_delaunay_triangulation():
    # all labels different
    points = [(50,50),(25,100),(50,150),(150,60),(150,125)]
    edges = delaunay_from_points(points,range(len(points)))
    assert len(edges) == 7
    assert [0,1] in edges
    assert [0,2] in edges
    assert [0,3] in edges
    assert [0,4] in edges
    assert [1,2] in edges
    assert [2,4] in edges
    assert [3,4] in edges
    # some doublettes
    points = [(50,50),(25,100),(50,150),(150,60),(150,125)]
    labels = [0,1,2,3,3]
    edges = delaunay_from_points(points,labels)
    assert len(edges) == 5
    assert [0,1] in edges
    assert [0,2] in edges
    assert [0,3] in edges
    assert [1,2] in edges
    assert [2,3] in edges
    # collinear edge resolution
    points = [(50,50),(50,100),(50,150),(50,200),(150,125)]
    edges = delaunay_from_points(points,range(len(points)))
    assert len(edges) == 7
    assert [0, 1] in edges
    assert [0, 4] in edges
    assert [1, 2] in edges
    assert [1, 4] in edges
    assert [2, 3] in edges
    assert [2, 4] in edges
    assert [3, 4] in edges



#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom,
#                          and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#


from gamera.plugin import * 
import re, string


def build_id_regex(s):
    def _build_id_regex_parts(s):
        if s == '':
            return ''
        regex_parts = []
        parts = s.split('.')
        for part in parts:
            ors = part.split('|')
            l = []
            for part0 in ors:
                part0 = part0.replace('*', '[^.]*')
                part0 = part0.replace('?', '[^.]')
                l.append(part0)
            regex_parts.append('(?:%s)' % '|'.join(['(?:%s)' % x for x in l]))
        regex = '\.'.join(regex_parts)
        return regex

    def _build_id_regex_parens(s):
        lparen = s.find('(')
        rparen = s.find(')')
        if s == '':
            return '', ''
        elif s[0] == '|':
            r, s = _build_id_regex_parens(s[1:])
            return '|' + r, s
        elif lparen != -1 and lparen < rparen:
            result = _build_id_regex_parts(s[:lparen])
            r1, s1 = _build_id_regex_parens(s[lparen+1:])
            r2, s2 = _build_id_regex_parens(s1)
            return result + r1 + r2, s2
        elif rparen != -1:
            return _build_id_regex_parts(s[:rparen]), s[rparen+1:]
        else:
            return _build_id_regex_parts(s), ''
    regex, s = _build_id_regex_parens(s)
    return re.compile(regex, re.IGNORECASE)

regex_cache = {}
dummy_regex = re.compile('')
type_dummy_regex = type(dummy_regex)
class match_id_name(PluginFunction):
    r"""
    Returns true if the image's main ``id_name`` matches the given regular
    expression.

    *regex*
      A special-purpose regular expression as defined below.

    A special-purpose regular expression language designed for
    convenience on Gamera symbol names is used here.  It has limited
    flexibility relative to Python regular expressions, and is more
    akin to wildcards in command-line shells.  This language is
    converted to standard Python regular expressions (``re`` module)
    for efficiency.

    **Informal syntax definition:**

    ====================  =============================================
    Regular expression    Description
    ====================  =============================================
    ``A|B``               matches ``A`` or ``B``
    ``A.B|C``             matches ``A.B`` or ``A.C``
    ``*``                 multiple-character wildcard
    ``?``                 single character wildcard
    ``()``                grouping can be performed with parentheses
    ``[a-z]``             matches any character a-z
    ====================  =============================================
    
    **Example expressions:**
    
    ========================  ====================================================  
    ``(upper.x)|(lower.y)``   match either ``upper.x`` or ``lower.y``
    ``upper.x|upper.y``       careful! matches ``upper.x.y`` or ``upper.upper.y``
    ``upper.*``               match anything in the ``upper`` category
    ``upper.a|b|c``           matches ``upper.a``, ``upper.b`` or ``upper.c``
    ``upper.capital_?``       ``?`` is a single character wildcard
    ========================  ====================================================
    """
    args = Args([String('regex')])
    def __call__(self, regex):
        global regex_cache
        if type(regex) == StringType:
            compiled = regex_cache.get(regex, None)
            if compiled is None:
                compiled = build_id_regex(regex)
                regex_cache[regex] = compiled
        elif type(regex) == type_dummy_regex:
            compiled = regex
        else:
            raise TypeError(
                "Expression must be a string or a compiled Python regular expression.")
        name = self.get_main_id()
        if compiled.match(name):
            return 1
        #for confidence, name in self.id_name:
        #    if compiled.match(name):
        #        return 1
        return 0
    __call__ = staticmethod(__call__)

_valid = string.letters + string.digits + "_"
def id_name_to_identifier(symbol):
    while len(symbol) and symbol[0] == '.':
        symbol = symbol[1:]
    if symbol == '':
        return '____'
    def translate(c):
        if not c in _valid:
            return "_"
        return c
    symbol = ''.join([translate(x) for x in symbol])
    symbol = symbol.lower()
    if symbol[0] in string.digits or keyword.iskeyword(symbol):
        symbol = '_' + symbol + '_'
    # Split by '.' delimiters
    orig_tokens = symbol.strip().split('.')
    tokens = []
    for token in orig_tokens:
        if token.strip() != '':
            tokens.append(token.strip())
    symbol = '_'.join(tokens)
    return symbol

class IdNameMatchingModule(PluginModule):
    category = "Classification"
    functions = [match_id_name]
    author = "Michael Droettboom and Karl MacMillan"
    url = "http://gamera.sourceforge.net/"
    pure_python = 1

module = IdNameMatchingModule()

match_id_name = match_id_name()



#!/usr/bin/env python
# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom, Karl MacMillan
#               2016      Christoph Dalitz
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

######################################################################
# Import Python stdlib
from __future__ import generators
import codecs, locale, sys, glob, cStringIO, inspect, getopt, os.path, shutil
from stat import ST_MTIME
from time import strftime, localtime
import locale
import traceback
import warnings
from distutils.version import StrictVersion, LooseVersion
from gamera.util import dedent
try:
   locale.setlocale(locale.LC_ALL, '')
except Exception:
   pass

######################################################################
# Import docutils
try:
   from docutils.core import publish_file
   import docutils.parsers.rst
   if StrictVersion(docutils.__version__) < StrictVersion('0.4'):
      print "docutils version (" + docutils.__version__ + ") too old"
      raise ImportError()
except ImportError, e:
   print "'docutils' 0.4 or later must be installed to generate the documentation."
   print "It can be downloaded at http://docutils.sf.net"
   sys.exit(1)

######################################################################
# Import pygments or SilverCity
source_highlighter = None
try:
   import pygments
   import pygments.lexers
   import pygments.formatters
   source_highlighter = 'pygments'
   if LooseVersion(pygments.__version__) < LooseVersion('0.6'):
      print "pygments version (" + pygments.__version__ + ") too old"
      raise ImportError()
except ImportError, e:
   try:
      import SilverCity
      source_highlighter = 'silvercity'
   except ImportError, e:
      print "Either 'pygments' 0.6 or later or 'SilverCity' 0.9 or later"
      print "must be installed to colorize the sourcecode snippets in the"
      print "documentation."
      sys.exit(1)

# Source code highlighting support
if source_highlighter == 'pygments':
   html_formatter = pygments.formatters.HtmlFormatter()

   def code_block(name, arguments, options, content, lineno,
                  content_offset, block_text, state, state_machine):
      language = arguments[0].lower()
      try:
         lexer = pygments.lexers.get_lexer_by_name(language)
      except ValueError:
         # no lexer found - use the text one instead of an exception
         error = state_machine.reporter.error(
            "No pygments lexer found for language '%s'." % language, 
            docutils.nodes.literal_block(block_text, block_text), line=lineno)
         return [error]
      parsed = pygments.highlight(
         u'\n'.join(content), 
         lexer, 
         html_formatter)
      return [docutils.nodes.raw('', parsed, format='html')]
elif source_highlighter == 'silvercity':
   def code_block(name, arguments, options, content, lineno,
                  content_offset, block_text, state, state_machine):
      language = arguments[0]
      try:
         module = getattr(SilverCity, language)
         generator = getattr(module, language+"HTMLGenerator")
      except AttributeError:
         error = state_machine.reporter.error(
            "No SilverCity lexer found for language '%s'." % language, 
            docutils.nodes.literal_block(block_text, block_text), line=lineno)
         return [error]
      io = cStringIO.StringIO()
      generator().generate_html( io, '\n'.join(content) )
      html = u'<div class="code-block">\n%s\n</div>\n' % io.getvalue()
      raw = docutils.nodes.raw('',html, format = 'html')
      return [raw]
code_block.arguments = (1,0,0)
code_block.options = {'language' : docutils.parsers.rst.directives.unchanged}
code_block.content = 1
docutils.parsers.rst.directives.register_directive( 'code', code_block )

######################################################################
# Import Gamera

try:
   from gamera import core, args, paths, util, plugin
   from gamera.plugins import png_support, tiff_support, _png_support
   from gamera.plugins import _image_conversion
   from gamera.enums import *
except ImportError, e:
   print "Cannot load gameracore."
   print "Gamera must be built before you can generate the documentation."
   sys.exit(1)

######################################################################

def sort_lowercase(a, b):
   return cmp(a.lower(), b.lower())

def sort_firstitem_lowercase(a, b):
   return cmp(a[0].lower(), b[0].lower())

_underline_levels = "=-`:'"
def underline(level, s, extra=0):
   return _underline_levels[level] * (len(s) + extra)

class DocumentationGenerator:
   def __init__(self, root_path=".", test_mode=False, classes=[], plugins=None, sourceforge_logo=False, contact_url="http://gamera.informatik.hsnr.de/contact.html"):
      self.set_paths(root_path)
      self.test_mode = test_mode
      self.classes = classes
      self.plugins = plugins
      self.sourceforge_logo = sourceforge_logo
      self.contact_url = contact_url

   def set_paths(self, root):
      def check_path(path):
         if not os.path.exists(path):
            raise RuntimeError(u"Documentation directory '%s' does not exist" % os.path.abspath(path))
         
      join = os.path.join
      self.root_path = root
      check_path(self.root_path)
      self.src_path = join(root, "src/")
      check_path(self.src_path)
      self.src_images_path = join(self.src_path, "images/")
      check_path(self.src_images_path)
      self.icons_path = join(root, "../gamera/pixmaps/")
      self.output_path = join(self.root_path, "html/")
      self.output_images_path = join(self.output_path, "images/")
      if not os.path.exists(self.output_path):
         os.mkdir(self.output_path)
      if not os.path.exists(self.output_images_path):
         os.mkdir(self.output_images_path)

   def generate(self):
      self.copy_css(self.src_path, self.output_path)
      self.generate_generic_pngs()
      PluginDocumentationGenerator(self, self.plugins)
      ClassDocumentationGenerator(self, self.classes)
      self.copy_images([self.src_images_path, self.icons_path],
                       self.output_images_path)
      if not self.test_mode:
         self.convert_to_html()
      print

   def generate_generic_pngs(self):
      print "Copying over generic images"
      for pixel_type in ALL:
         if pixel_type in (FLOAT, COMPLEX):
            pixel_type_name = "GreyScale"
         else:
            pixel_type_name = util.get_pixel_type_name(pixel_type)
         image = core.load_image(
            os.path.join(paths.test,
                         pixel_type_name + "_generic.tiff"))
         print "  " + image.pixel_type_name
         _png_support.save_PNG(
            image,
            os.path.join(self.output_images_path,
                         "%s_generic.png" % (pixel_type_name)))

   def copy_css(self, input_path, output_path):
      print "Copying CSS file"
      for filename in ['default.css', 'html4css1.css']:
         shutil.copyfile(os.path.join(input_path, filename),
                         os.path.join(output_path, filename))
      if source_highlighter == 'pygments':
         fd = open(os.path.join(output_path, 'pygments.css'), 'w')
         fd.write(html_formatter.get_style_defs(""))
         fd.close()

   def copy_images(self, src_paths, output_path):
      print "Copying images"
      for path in src_paths:
         if os.path.exists(path):
            for file in glob.glob(os.path.join(path, "*.*")):
               path, filename = os.path.split(file)
               shutil.copyfile(file, os.path.join(output_path, filename))

   def get_rest_docs(self):
      locale.setlocale(locale.LC_TIME, 'C')
      for file in glob.glob(self.src_path + "*.txt"):
         path, filename = os.path.split(file)
         root, ext = os.path.splitext(filename)
         yield (file, root,
                codecs.open(file, 'r', 'utf-8'),
                strftime("%B %d, %Y", localtime(os.stat(file)[ST_MTIME])))

   def convert_to_html(self):
      print "Converting to HTML"
      for filename, rootname, fd, mtime in self.get_rest_docs():
         output_file = os.path.join(self.output_path, rootname + ".html")
         if (not os.path.exists(output_file) or
             os.stat(filename)[ST_MTIME] > os.stat(output_file)[ST_MTIME]):
            print "  Generating " + rootname
            lines = fd.readlines()
            if self.sourceforge_logo:
               footer = '.. footer:: :raw-html:`<a href="http://sourceforge.net/projects/gamera" style="float:left;"><img src="http://sflogo.sourceforge.net/sflogo.php?group_id=99328&type=13" width="120" height="30" border="0" alt="Get Gamera at SourceForge.net. Fast, secure and Free Open Source software downloads" /></a> <div style="text-align:right;">For contact information, see <a href="' + self.contact_url + '">' + self.contact_url + '</a></div>`\n\n'
            else:
               footer = '.. footer:: :raw-html:`<div style="text-align:right;">For contact information, see <a href="' + self.contact_url + '">' + self.contact_url + '</a></div>`\n\n'
            lines = (lines[:3] + 
                     ["\n", u"**Last modified**: %s\n\n" % mtime, 
                      ".. contents::\n\n", 
                      ".. role:: raw-html(raw)\n   :format: html\n",
                      footer
                      ] + 
                     lines[3:])
            fd = cStringIO.StringIO((u''.join(lines)).encode("utf-8"))
            try:
               overrides = {'embed_stylesheet': False,
                            'stylesheet_path': None,
                            'stylesheet': 'default.css'}
               try:
                  publish_file(source=fd, source_path=filename, destination_path=output_file, 
                               writer_name="html", settings_overrides=overrides)
               except TypeError:
                  publish_file(source=fd, destination_path=output_file)
            except KeyboardInterrupt, e:
               raise e
            except Exception, e:
               traceback.print_exc()
               print "Unable to convert '%s' to HTML: %s" % (filename, e)
               if os.path.exists(output_file):
                  os.remove(output_file)

class PluginDocumentationGenerator:
   def __init__(self, docgen, plugins=None):
      print "Generating plugin documentation"

      self.docgen = docgen

      images = self.get_generic_images()
      images['src_images_path'] = docgen.src_images_path

      methods = self.get_methods(plugins)
      if len(methods) == 0:
         return

      self.recurse(methods, 0, images)
      self.table_of_contents(methods)
      
   def get_generic_images():
      images = {}
      for i in [ONEBIT, RGB, GREYSCALE, GREY16]:
         pixel_type_name = util.get_pixel_type_name(i)
         images[i] = core.load_image(os.path.join(
            paths.test, pixel_type_name + "_generic.tiff"))
      for i in (FLOAT, COMPLEX):
         image = core.load_image(os.path.join(
            paths.test, "GreyScale_generic.tiff"))
         if i == FLOAT:
            images[i] = _image_conversion.to_float(image)
         elif i == COMPLEX:
            try:
               images[i] = _image_conversion.to_complex(image)
            except Exception:
               pass
      return images
   get_generic_images = staticmethod(get_generic_images)

   def get_methods(plugins):
      def methods_flatten(dest, source, flat, all=False):
         for key, val in source.items():
          if type(val) == dict:
            if key != "Test":
               if plugins is None or key in plugins or all:
                  if not dest.has_key(key):
                     dest[key] = {}
                  methods_flatten(dest[key], val, flat, True)
          else:
            dest[key] = val
            flat[key] = val

      methods = plugin.plugin_methods
      flat_methods = {}
      flat_list = {}
      for pixel_type in ALL + [NONIMAGE]:
         if methods.has_key(pixel_type): 
             methods_flatten(flat_methods, methods[pixel_type],
                             flat_list, False)

      return flat_methods
   get_methods = staticmethod(get_methods)

   def recurse(self, methods, level, images, s=None):
      methods_list = methods.items()
      methods_list.sort()
      for key, val in methods_list:
         if type(val) == dict:
            if level == 0:
                filename = key.lower()
                s = codecs.open(
                   os.path.join(self.docgen.src_path, filename + ".txt"),
                   "w", "utf-8")
            s.write(u"\n%s\n%s\n\n" % (key, underline(level, key)))
            print "  " * (level + 1) + key
            self.recurse(val, level + 1, images, s)
         else:
            self.method_doc(val, level, images, s)

   def table_of_contents(self, methods):
      def toc_recurse(s, methods, level, links, index, filename=None):
          methods_list = methods.items()
          methods_list.sort()
          for key, val in methods_list:
             if key.startswith("_"):
                key = key[1:]
             if type(val) == dict and level == 0:
                 filename = key.lower()
             s.write("  " * level)
             s.write("- ")
             s.write(key)
             s.write("_")
             s.write("\n\n")
             bookmark = key.lower().replace("_", "-")
             if bookmark.startswith("-"):
                bookmark = bookmark[1:]
             href = "%s.html#%s" % (filename, bookmark)
             links.append(".. _%s: %s" % (key, href))
             if type(val) == dict:
                 toc_recurse(s, val, level + 1, links, index, filename)
             else:
                 index.append(key)

      s = codecs.open(os.path.join(self.docgen.src_path, "plugins.txt"),
                      "w", "utf-8")
      s.write("=======\nPlugins\n=======\n\nBy categories\n-------------\n\n")
      links = []
      index = []
      toc_recurse(s, methods, 0, links, index)
      s.write("Alphabetical\n-------------\n")
      index.sort(sort_lowercase)
      letter = ord('A') - 1
      first = True
      for name in index:
          if ord(name[0].upper()) > letter:
              letter = ord(name[0].upper())
              s.write("\n\n**")
              s.write(chr(letter))
              s.write("**\n\n")
              first = True
          if not first:
              s.write(", ")
          s.write(name)
          s.write("_")
          first = False
      s.write("\n\n")
      s.write("\n".join(links))
      
   def method_doc(self, func, level, images, s):
      s.write(u"``%s``\n" % func.__name__)
      s.write(underline(level, func.__name__, 4))
      s.write("\n\n")
      if func.return_type != None:
         s.write(func.return_type.rest_repr() + " ")
      header = u"**%s** (%s)\n" % (func.__name__, ', '.join(
          [x.rest_repr(True) for x in func.args.list]))
      s.write(header)
      s.write("\n\n")
      if func.self_type != None:
         s.write(u":Operates on: %s\n" % func.self_type.rest_repr(False))
      if func.return_type != None:
         s.write(u":Returns: %s\n" % (func.return_type.rest_repr(False)))
      if func.category == None:
         category = func.module.category
      else:
         category = func.category
      if category != None:
         s.write(u":Category: %s\n" % category)
      file = os.path.split(inspect.getsourcefile(func))[1]
      if file == 'plugin.py':
          file = 'core.py'
      s.write(":Defined in: %s\n" % os.path.split(inspect.getsourcefile(func))[1])
      if func.author is None:
          author = func.module.author
      else:
          author = func.author
      if author != None:
         if func.module.url != None and 0:
            s.write(u":Author: `%s`__\n.. __: %s\n" % (author, func.module.url))
         else:
            s.write(u":Author: %s\n" % (author,))
      if func.image_types_must_match:
         s.write("\n*All images passed in (including self) must have the same pixel type.*\n\n")
      doc = func.__doc__
      if doc is None or doc == "":
         s.write("\n.. warning:: No documentation written.\n\n")
      else:
         doc = util.dedent(doc)
         s.write(u"\n\n%s\n" % doc)
      self.method_example(func, level, images, s)

   def run_example(self, func, images):
      results = []
      for doc_example in func.doc_examples:
         src_image = None
         pixel_type = None
         arguments = None
         display_arguments = None
         if inspect.isroutine(doc_example):
            result = doc_example(images)
         else:
            if len(doc_example):
               if isinstance(func.self_type, args.ImageType):
                  pixel_type = doc_example[0]
                  src_image = images[pixel_type].image_copy()
                  arguments = [src_image] + list(doc_example[1:])
                  display_arguments = doc_example[1:]
               else:
                  display_arguments = arguments = doc_example
            else:
               display_arguments = arguments = []
            result = func.__call__(*tuple(arguments))
         # add quotes around string args for display
         displayargs = None
         if display_arguments != None:
            displayargs = []
            for da in display_arguments:
               if isinstance(da,str):
                  displayargs.append("\"" + da + "\"")
               else:
                  displayargs.append(da)
         results.append((result, src_image, pixel_type, displayargs))
      return results

   def method_example(self, func, level, images, s):
      results = self.run_example(func, images)
      if len(results):
         s.write("\n----------\n\n")
      for i, (result, src_image, pixel_type, arguments) in enumerate(results):
         s.write(u"**Example %d:** %s" % (i + 1, func.__name__))
         if not arguments is None:
            s.write(u"(%s)" % ", ".join([str(x) for x in arguments]))
         s.write("\n\n")
         if not pixel_type is None:
            pixel_type_name = util.get_pixel_type_name(pixel_type)
            if pixel_type in (FLOAT, COMPLEX):
               pixel_type_name = "GreyScale"
            self.write_image(s, "%s_generic" % pixel_type_name)
         result_filename = "%s_plugin_%02d" % (func.__name__, i)
         if isinstance(result, core.ImageBase):
            self.save_image(result, result_filename)
            self.write_image(s, result_filename)
         elif (result is None and
               isinstance(func.self_type, args.ImageType) and
               src_image is not None):
            self.save_image(src_image, result_filename)
            self.write_image(s, result_filename)
         elif util.is_image_list(result):
            result_filenames = []
            for j, part in enumerate(result):
               result_filename = ("%s_plugin_%02d_%02d" %
                                  (func.__name__, i, j))
               self.save_image(part, result_filename)
               s.write("|%s| " % result_filename)
               result_filenames.append(result_filename)
            s.write("\n\n")
            for result_filename in result_filenames:
               self.write_image(s, result_filename, "|%s|" % result_filename)
         else:
            s.write("*result* = " + repr(result) + "\n\n")
      s.write("\n\n")

   def save_image(self, image, filename):
      if image.data.pixel_type in (COMPLEX, FLOAT):
         image = _image_conversion.to_greyscale(image)
      _png_support.save_PNG(
         image,
         os.path.join(
         self.docgen.output_images_path, filename + ".png"))

   def write_image(self, s, filename, tag=""):
      image = _png_support.load_PNG(os.path.join(self.docgen.output_images_path, filename + ".png"), 0)
      s.write(".. %s image:: images/%s.png\n   :height: %d\n   :width: %d\n\n" %
              (tag, filename, image.height, image.width))


class ClassDocumentationGenerator:
   def __init__(self, docgen, classes):
      if not len(classes):
         return
      print "Generating class documentation"
      self.docgen = docgen
      self.class_names = []
      for cls in classes:
         self.document_class(cls)
      self.table_of_contents(classes)

   def document_class(self, cls_desc):
      module_name, cls_name, members = cls_desc
      combined_name = ".".join([module_name, cls_name])
      self.class_names.append((cls_name, combined_name))
      print "  ", combined_name
      if type(members) in (str, unicode):
         members = members.split()
      s = codecs.open(os.path.join(self.docgen.src_path, combined_name + ".txt"), "w", "utf-8")
      s.write(u"class ``%s``\n%s\n\n" % (cls_name, underline(0, cls_name, 10)))
      s.write(u"``%s``\n%s\n\nIn module ``%s``\n\n" % (cls_name, underline(1, cls_name, 4), module_name))
      s.write(u".. docstring:: %s %s\n   :no_title:\n\n" % (module_name, cls_name))
      s.write(u".. docstring:: %s %s %s\n\n" % (module_name, cls_name, " ".join(members)))

   def table_of_contents(self, classes):
      s = codecs.open(os.path.join(self.docgen.src_path, "classes.txt"), "w", "utf-8")
      s.write("=======\nClasses\n=======\n\n")
      s.write("Alphabetical\n-------------\n")
      self.class_names.sort(sort_firstitem_lowercase)
      letter = '~'
      first = True
      links = []
      for name, combined_name in self.class_names:
          if name[0].upper() != letter:
              letter = name[0].upper()
              s.write("\n\n**")
              s.write(letter)
              s.write("**\n\n")
              first = True
          if not first:
              s.write(", ")
          s.write(u"%s_ (%s)" % (name, combined_name))
          first = False
          links.append(".. _%s: %s.html" % (name, combined_name))
      s.write("\n\n")
      s.write("\n".join(links))
      s.close()

# For extracting docstrings
def docstring(name, arguments, options, content, lineno,
              content_offset, block_text, state, state_machine):
   def import_helper(module_name, object_name):
      mod = __import__(module_name)
      components = module_name.split(".")[1:] + object_name.split(".")
      for comp in components:
         mod = getattr(mod, comp)
      return mod

   def do_docstring(obj, name):
      doc_string = inspect.getdoc(obj)
      if doc_string is None:
         text = "[Not documented]\n\n"
      else:
         text = doc_string + "\n\n"
      content = docutils.statemachine.string2lines(text, convert_whitespace=1)
      if not options.has_key('no_title'):
         title_text = u"``%s``" % name
         textnodes, messages = state.inline_text(title_text, lineno)
         titles = [docutils.nodes.title(title_text, '', *textnodes)] + messages
         node = docutils.nodes.section(text, *titles)
         node['names'] = [name]
         node['name'] = name
         state_machine.document.note_implicit_target(node)
      else:
         node = docutils.nodes.paragraph(text)
      content = docutils.statemachine.StringList(initlist=content, parent=node)
      state.nested_parse(content, content_offset, node)
      return node

   base_obj = import_helper(*arguments[:2])
   if len(arguments) == 2:
      name = arguments[1].split(".")[-1]
      return [do_docstring(base_obj, name)]
   nodes = []
   for name in arguments[2:]:
      try:
         obj = getattr(base_obj, name)
      except AttributeError:
         print "Warning: '%s' could not be found in '%s'" % (name, base_obj.__name__)
      else:
         nodes.append(do_docstring(obj, name))
   return nodes
docstring.arguments = (2,256,0)
docstring.options = { "no_title": docutils.parsers.rst.directives.flag }
docstring.content = 0
docutils.parsers.rst.directives.register_directive( 'docstring', docstring)

def copy_images(path_obj):
   if not os.path.exists(path_obj.output_images_path):
      os.mkdir(path_obj.output_images_path)
   for path in (path_obj.doc_images_path,
                path_obj.src_images_path,
                path_obj.icons_path):
      for file in glob.glob(path + "*.*"):
         path, filename = os.path.split(file)
         open(os.path.join(path_obj.output_images_path, filename), "wb").write(
             open(file, "rb").read())

class Paths:
    def __init__(self, root="."):
        join = os.path.join
        self.root_path = root
        self.src_path = join(root, "src/")
        self.src_images_path = join(self.src_path, "images/")
        self.icons_path = join(root, "../gamera/pixmaps/")
        self.output_path = join(self.root_path, "html/")
        self.output_images_path = join(self.output_path, "images/")

def print_usage():
    print "Gamera documentation generator."
    print "Usage: gendoc.py [-d doc_directory]"
    print "  (if doc_directory is omitted, the current directory"
    print "   will be used.)"
    print 
   
def gendoc(classes=[], plugins=None, sourceforge_logo=False, contact_url="http://gamera.informatik.hsnr.de/contact.html"):
   print_usage()
   opts, args = getopt.getopt(sys.argv[1:], "d:t")
   root = '.'
   test_mode = False
   for flag, value in opts:
      if flag == "-d":
         root = value
      elif flag == "-t":
         test_mode = True
   try:
      docgen = DocumentationGenerator(root, test_mode, classes=classes, plugins=plugins, sourceforge_logo=sourceforge_logo, contact_url=contact_url)
   except Exception, e:
      print "Documentation generation failed with the following exception:"
      print e
   else:
      try:
         docgen.generate()
      except KeyboardInterrupt:
         print "Documentation generation stopped by user."



#
# Copyright (C) 2013      Christian Brandt and Christoph Dalitz
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from gamera.plugin import *
import _fourier_features

class fourier_broken(PluginFunction):
	"""Fourier descriptor for arbitrary (not necessarily connected) shapes
according to equations (18) and (19) in 

  C. Dalitz, C. Brandt, S. Goebbels, D. Kolanus:
  `Fourier Descriptors for Broken Shapes`__.
  EURASIP Journal on Advances in Signal Processing 2013:161, 2013
 
The coefficient c_0 is used for scale normalisation. The 
absolute values of the coefficients A(0), A(N-1) A(1), A(N-2), ... are 
returned.

    +---------------------------+
    | **Invariant to:**         |  
    +-------+----------+--------+
    | scale | rotation | mirror | 
    +-------+----------+--------+
    |   X   |     X    |        |
    +-------+----------+--------+

.. __: http://dx.doi.org/10.1186/1687-6180-2013-161
"""
	category = "Features"
	self_type = ImageType([ONEBIT])
	return_type = FloatVector(length=48)
	feature_function = True
        doc_examples = [(ONEBIT,)]

class FourierFeaturesModule(PluginModule):
	cpp_headers = ["fourier_features.hpp"]
	category = "Features"
	functions = [fourier_broken]
	cpp_sources = ["src/geostructs/kdtree.cpp", "src/geostructs/delaunaytree.cpp"]
	extra_compile_args = ["-DFDLENGTH=48"]
	author = "Christian Brandt and Christoph Dalitz"
	url = "http://gamera.sourceforge.net/"

module = FourierFeaturesModule()



#!/usr/bin/python

#
# Copyright (C) 2008-2009 Michael Droettboom
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
#

# This is a total hack to generate and upload the daily builds.
# Nothing about this file should be expected to be portable between
# machines.

# This script runs successfully on an Ubuntu 8.04 LTS system with
# (at least) the following packages installed:
#     python-dev
#     python2.4
#     python2.4-dev
#     mingw32
#     g++
#     libpng12-dev
#     libtiff4-dev
#     python-docutils
#     python-pygments
#     openssh-client
#     rsync

import datetime
import os
import shutil
import sys
from distutils.util import get_platform
import subprocess

import pysvn

ROOT_PATH = "/home/mdboom/JHU/builds/"
WORKING_PATH = "gamera-daily-build"
REPOS_PATH = "https://gamera.svn.sf.net/svnroot/gamera/trunk/gamera"
STAGING_PATH = "gamera-daily"
RSYNC_TARGET = "mdboom@xs004.mse.jhu.edu:/mnt/gamera-builds/"
KEEP = 10

def mysystem(message, command):
    print message,
    try:
        # process = subprocess.Popen(command, shell=True,
        #                            stdout=subprocess.PIPE,
        #                            stderr=subprocess.PIPE)
        # retcode = process.wait()
        retcode = os.system(command)
        if retcode != 0:
            print "FAILED (%d)" % retcode
            return
    except Exception:
        print "FAILED"
        raise
    else:
        print "SUCCESS"

def myrmtree(path):
    if os.path.exists(path):
        shutil.rmtree(path)

def mymkdir(path):
    if not os.path.exists(path):
        os.mkdir(path)

def update_working_copy():
    client = pysvn.Client()

    push_update = False
    if os.path.exists(WORKING_PATH):
        current_working = client.info(WORKING_PATH)['revision'].number
        current_remote = 0
        for rev in client.update(WORKING_PATH):
            current_remote = max(current_remote, rev.number)
        if current_remote > current_working:
            print "Updates in SVN"
            push_update = True
    else:
        print "Checking out clean"
        client.checkout(REPOS_PATH, WORKING_PATH)
        current_working = client.info(WORKING_PATH)['revision'].number
        push_update = True
    return push_update, current_working

def get_version(now, current_working):
    return "daily-%s-SVN-r%s" % (now, current_working)

def update_version(version):
    fd = open(os.path.join(WORKING_PATH, "version"), "w")
    fd.write(version)
    fd.close()

def build(version):
    build_dir = os.path.abspath(os.path.join(WORKING_PATH, "build", "lib.%s-%s" % (get_platform(), sys.version[0:3])))
    if "--clean" in sys.argv:
        myrmtree(os.path.join(WORKING_PATH, "build"))
        myrmtree(os.path.join(WORKING_PATH, "dist"))
    myrmtree(os.path.join(WORKING_PATH, "doc/html"))
    os.chdir(WORKING_PATH)
    mysystem("Building for Linux (Python 2.6)...", "python setup.py build")
    mysystem("Building source distribution...", "python setup.py sdist")
    mysystem("Building for Windows...", "python setup.py build --compiler=mingw32_cross bdist_wininst")
    os.chdir("doc")
    mysystem("Building docs...", "export PYTHONPATH=%s:$PYTHONPATH; python gendoc.py" % build_dir)
    os.rename("html", "gamera-doc-%s" % version)
    mysystem("tar-gzipping docs...", "tar czvf gamera-doc-%s.tar.gz gamera-doc-%s" % (version, version))
    mysystem("zipping docs...", "zip -r gamera-doc-%s.zip gamera-doc-%s" % (version, version))
    os.rename("gamera-doc-%s" % version, "html")
    os.chdir("../..")

def rotate(path, number=2):
    files = []
    for f in os.listdir(path):
        p = os.path.join(path, f)
        files.append((os.stat(p).st_mtime, p))
    if len(files) < KEEP * number:
        return
    files.sort()
    files.reverse()
    remove = files[KEEP * number:]
    for t, f in remove:
        os.path.remove(f)

def stage(version):
    mymkdir(os.path.join(STAGING_PATH, "src"))
    mymkdir(os.path.join(STAGING_PATH, "win32"))
    mymkdir(os.path.join(STAGING_PATH, "doc"))
    shutil.copy2(os.path.join(WORKING_PATH, "dist", "gamera-%s.tar.gz" % version),
                 os.path.join(STAGING_PATH, "src"))
    shutil.copy2(os.path.join(WORKING_PATH, "dist", "gamera-%s.win32-py%d.%d.exe" %
                              (version, sys.version_info[0], sys.version_info[1])),
                 os.path.join(STAGING_PATH, "win32"))
    shutil.copy2(os.path.join(WORKING_PATH, "doc", "gamera-doc-%s.tar.gz" % version),
                 os.path.join(STAGING_PATH, "doc"))
    shutil.copy2(os.path.join(WORKING_PATH, "doc", "gamera-doc-%s.zip" % version),
                 os.path.join(STAGING_PATH, "doc"))
    rotate(os.path.join(STAGING_PATH, "src"))
    rotate(os.path.join(STAGING_PATH, "win32"))
    rotate(os.path.join(STAGING_PATH, "doc"), 2)

def rsync():
    mysystem("uploading", "rsync -e ssh -arvz --delete %s/* %s" %
             (STAGING_PATH, RSYNC_TARGET))

def main():
    os.chdir(ROOT_PATH)
    now = datetime.datetime.now().strftime("%Y%m%d")
    mymkdir(STAGING_PATH)
    log_file = open(os.path.join(STAGING_PATH, "daily-build.log"), "a")
    sys.stdout = log_file
    sys.stderr = log_file
    print "-" * 76
    print "RUNNING:", now
    push_update, current_working = update_working_copy()
    if '--force' in sys.argv:
        push_update = True
    if not push_update:
        print "No updates today."
    else:
        version = get_version(now, current_working)
        update_version(version)
        build(version)
        stage(version)
    rsync()

if __name__ == '__main__':
    main()



#
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom, and Karl MacMillan
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from gamera.plugin import *
import _gui_support

class to_string(PluginFunction):
    """
    Encodes the image into a 'string' required by wxImage.
    (i.e. 8-bit RGB triplets).
    """
    self_type = ImageType([ONEBIT, GREYSCALE, GREY16, RGB, FLOAT, COMPLEX])
    return_type = Class("image_as_string")

class to_buffer(PluginFunction):
    """
    Encodes the image into a 'buffer' required by wx.Image.
    (i.e. 8-bit RGB triplets). If you need to convert a gamera image
    *scaled_image* to a wx.Bitmap, you can do so as follows:

    .. code:: Python
        
      wximg = wx.EmptyImage(scaled_image.ncols, scaled_image.nrows)
      scaled_image.to_buffer(wximg.GetDataBuffer())
      wxbmp = wx.BitmapFromImage(wximg)
    """
    category = "ExternalLibraries"
    self_type = ImageType([ONEBIT, GREYSCALE, GREY16, RGB, FLOAT, COMPLEX])
    args = Args(Class("Buffer"))

class draw_cc(PluginFunction):
    """
    Draws a colored Cc over an already initialized wxImage buffer.
    """
    self_type = ImageType([RGB])
    args = Args([ImageType([ONEBIT]),
                 Int("red"), Int("green"), Int("blue")])

class to_buffer_colorize(PluginFunction):
    """
    Encodes the image into a 'buffer' required by wxImage, and
    applies the given color to the foreground and background.
    """
    self_type = ImageType([ONEBIT, GREYSCALE])
    args = Args([Class("Buffer"),
                 Int("red"), Int("green"), Int("blue"),
                 Bool("invert")])

class color_ccs(PluginFunction):
    """
    Returns an RGB image where each connected component of the
    image is colored one of eight different colors.  This function can
    be used to verify that ``cc_analysis`` is working correctly for your
    image.

    It should be noted that this coloring does not take care of
    component adjacency. It may therefore happen that adjacent components
    obtain the same color. If this is not acceptable in your use case,
    have a look at graph_color_ccs_ instead.

    .. _graph_color_ccs: color.html#graph-color-ccs

    *ignore_unlabeled*:
      do not colorize unlabeled pixels (pixel value one), but leave them black

    .. note:: Connected component analysis must already be performed
              on the image (using cc_analysis_, for example) in order
              for this to work.
          
    .. _cc_analysis: segmentation.html#cc-analysis

    **Example 1:**

      .. image:: images/color_ccs.png
    """
    author = "Michael Droettboom, Karl MacMillan, and Robert Butz"
    args = Args([Check('ignore_unlabeled', default=True)])
    category = "Color"
    self_type = ImageType([ONEBIT])
    return_type = ImageType([RGB])
    def __call__(image, ignore_unlabeled=False):
        return _gui_support.color_ccs(image, ignore_unlabeled)
    __call__ = staticmethod(__call__)

# By default, the wxPython-devel RPM puts stuff here, but this
# should be done better
class GuiSupportModule(PluginModule):
    """
    This module provides various functions that support the GUI
    infrastructure.
    """
    category = None
    cpp_headers = ["gui_support.hpp"]
    functions = [to_string, to_buffer, to_buffer_colorize, color_ccs,
                 draw_cc]
    author = "Michael Droettboom and Karl MacMillan"
    url = "http://gamera.sourceforge.net/"
    
module = GuiSupportModule()



# -*- mode: python; indent-tabs-mode: nil; tab-width: 3 -*-
# vim: set tabstop=3 shiftwidth=3 expandtab:
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom, Karl MacMillan
#               2010-2012 Christoph Dalitz
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from gamera.plugin import *
import _misc_filters

class rank(PluginFunction):
  """
  Within each *k* times *k* window, set the center pixel to the *r*-th ranked
  value.

  Note that for ``Onebit`` images, actually *rank(k*k - r + 1)* is computed instead
  of *rank(r)*. This has the effect that you do not need to worry whether
  your image is a greyscale or onebit image: in all cases low values
  for *r* will darken the image and high values will light it up.

  *rank* (1, 2, ..., *k* * *k*)
    The rank of the windows pixels to select for the center. (k*k+1)/2 is
    equivalent to the median.

  *k* (3, 5 ,7, ...)
    The window size (must be odd).

  *border_treatment* (0, 1)
    When 0 ('padwhite'), window pixels outside the image are set to white.
    When 1 ('reflect'), reflecting boundary conditions are used.
  """
  self_type = ImageType([ONEBIT, GREYSCALE, GREY16, FLOAT])
  args = Args([Int('rank'), Int('k', default=3),
               Choice('border_treatment', ['padwhite', 'reflect'], default=1)])
  return_type = ImageType([ONEBIT, GREYSCALE, GREY16, FLOAT])
  author = "Christoph Dalitz and David Kolanus"
  doc_examples = [(GREYSCALE, 2), (GREYSCALE, 5), (GREYSCALE, 8)]
  def __call__(self, rank, k=3, border_treatment=1):
    if k%2 == 0:
      raise RuntimeError("rank: window size k must be odd")
    if rank < 1 or rank > k*k:
      raise RuntimeError("rank: rank must be between 1 and k*k")
    return _misc_filters.rank(self, rank, k, border_treatment)
  __call__ = staticmethod(__call__)

class mean(PluginFunction):
  """
  Within each *k* times *k* window, set the center pixel to the mean
  value of all pixels.

  *k* is the window size (must be odd), and *border_treatment* can
  be 0 ('padwhite'), which sets window pixels outside the image to white,
  or 1 ('reflect'), for reflecting boundary conditions.
  """
  self_type = ImageType([ONEBIT, GREYSCALE, GREY16, FLOAT])
  args = Args([Int('k', default=3),
               Choice('border_treatment', ['padwhite', 'reflect'], default=1)])
  doc_examples = [(GREYSCALE,)]
  return_type = ImageType([ONEBIT, GREYSCALE, GREY16, FLOAT])
  author = "David Kolanus"
  def __call__(self, k=3, border_treatment=1):
    if k%2 == 0:
      raise RuntimeError("mean: window size k must be odd")
    return _misc_filters.mean(self, k, border_treatment)
  __call__ = staticmethod(__call__)

class min_max_filter(PluginFunction):
    """
    Within each *k* times *k* window, set the center pixel to the minimum or
    maximum value of all pixels inside the window.

    *k* is the window size (must be odd) and *filter* is the filter type
    (0 for min, 1 for max). When *k_vertical* is nonzero, the vertical size of
    the window is set to *k_vertical* instead of *k*.

    This function does the same as *rank(1,k,border_treatment=1)*, but is
    much faster because the runtime of *min_max_filter* is constant in
    the window size. The same algorithm has been developed independently
    by van Herk and Gil and Werman. See

      M. van Herk: *A fast algorithm for local minimum and maximum filters
      on rectangular and octagonal kernels.* Pattern Recognition Letters 13,
      pp. 517-521, 1992

      J. Gil, M. Werman: *Computing 2-D min, median, and max filters.*
      IEEE Transactions on Pattern Analysis and Machine Intelligence 15,
      pp. 504-507, 1993
      """
    self_type = ImageType([ONEBIT, GREYSCALE, GREY16, FLOAT])
    args = Args([Int('k', default=3),
                 Choice('filter', ['min', 'max'], default=0),
                 Int('k_vertical', default=0)])
    return_type = ImageType([ONEBIT, GREYSCALE, GREY16, FLOAT])
    author = "David Kolanus"
    doc_examples = [(GREYSCALE,)]
    def __call__(self, k=3, filter=0, k_vertical=0):
        if k%2 == 0:
            raise RuntimeError("min_max_filter: window size k must be odd")
        if k_vertical != 0 and k_vertical%2 == 0:
            raise RuntimeError("min_max_filter: k_vertical must be zero or odd")
        return _misc_filters.min_max_filter(self, k, filter, k_vertical)
    __call__ = staticmethod(__call__)

class create_gabor_filter(PluginFunction):
    """
    Computes the convolution of an image with a two dimensional Gabor
    function. The result is returned as a float image.

    The *orientation* is given in radians, the other parameters are
    the center *frequency* (for example 0.375 or smaller) and the two
    angular and radial sigmas of the gabor filter.
    
    The energy of the filter is explicitly normalized to 1.0.
    """
    self_type = ImageType([GREYSCALE])
    return_type = ImageType([FLOAT])
    args = Args([Float("orientation", default=45.0),
                 Float("frequency", default=0.375),
                 Int("direction", default=5)])
    author = u"Ullrich K\u00f6the (wrapped from VIGRA by Uma Kompella)"
    doc_examples = [(GREYSCALE,)]
    def __call__(self,
                 orientation=45.0,
                 frequency=0.375,
                 direction=5):
        return _misc_filters.create_gabor_filter(self,
                                                 orientation,
                                                 frequency,
                                                 direction)
    __call__ = staticmethod(__call__)
    

class kfill(PluginFunction):
    """
    Removes salt and pepper noise in onebit images by applying the *kfill*
    filter *iterations* times. Should before *iteration* times a run of
    the kfill filter not change a single pixel, the iteration is stopped
    beforehand.

    In contrast to a rank or mean filter, kfill is designed in such
    a way that it does not merge non touching connected components.
    To this end, the border of the *k* times *k* mask is scanned for
    black pixels and the center is not filled when this would connect
    disjoint pixels on the border. A detailed description of the
    algorithm can be found in

       M. Seul, L. O'Gorman, M.J. Sammon: *Practical Algorithms for
       Image Analysis.* Cambridge University Press, 2000

    The present implementation does not use code from the book, but has
    been written from scratch.
    """
    self_type = ImageType([ONEBIT])
    return_type = ImageType([ONEBIT])
    author = "Oliver Christen"
    args = Args([Int("k", default=3),Int("iterations", default=1)])
    def __call__(self, k=3, iterations=1):
      if k < 3:
        raise RuntimeError("kfill: k must be >= 3")
      if iterations < 1:
        raise RuntimeError("kfill: number of iterations must be > 0")
      return _misc_filters.kfill(self, k, iterations)
    __call__ = staticmethod(__call__)

class kfill_modified(PluginFunction):
    """
    Removes salt and pepper noise in onebit images by applying a
    modified version of the *kfill* filter proposed in the following
    reference:

       K.Chinnasarn, Y.Rangsanseri, P.Thitimajshima:
       *Removing Salt-and-Pepper Noise in Text/Graphics Images.*
       Proceedings of The Asia-Pacific Conference on Circuits and Systems
       (APCCAS'98), pp. 459-462, 1998

    For *k* = 3, this algorithm is identical with the original kfill algorithm.
    For larger *k* however, it fills the window core also when not all
    pixels are of the same value. It should be noted that in this case,
    the modified version does *not* take care
    of connectivity. In other words, it can result in joining previously
    disconnected connected components, similar to a morphological closing
    operation, while at the same time small black speckles are removed.
    """
    self_type = ImageType([ONEBIT])
    return_type = ImageType([ONEBIT])
    author = "Oliver Christen"
    args = Args([Int("k", default=3)])
    def __call__(self, k=3):
    		if k < 3:
    			raise RuntimeError("k < 3")
    		return _misc_filters.kfill_modified(self, k)
    __call__ = staticmethod(__call__)


class MiscFiltersModule(PluginModule):
    category = "Filter"
    functions = [mean, rank, min_max_filter, create_gabor_filter,
                 kfill, kfill_modified]
    cpp_headers = ["misc_filters.hpp"]
    author = "Michael Droettboom and Karl MacMillan"
    url = "http://gamera.sourceforge.net/"
module = MiscFiltersModule()



#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom, Karl MacMillan
#               2009-2010 Christoph Dalitz
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import array
from gamera.plugin import *
import _features

class Feature(PluginFunction):
    self_type = ImageType([ONEBIT])
    return_type = FloatVector(length=1)
    feature_function = True
    doc_examples = [(ONEBIT,)]

class black_area(Feature):
    """
    The simplest of all feature-generating functions, ``black_area``
    simply returns the number of black pixels.

    +---------------------------+
    | **Invariant to:**         |  
    +-------+----------+--------+
    | scale | rotation | mirror | 
    +-------+----------+--------+
    |       |     X    |   X    |
    +-------+----------+--------+
    
    .. warning:: This feature is not scale invariant.
    """
    pass

class moments(Feature):
    """
    Returns *moments* of the image.

    The first two elements of the returned ``FloatVector`` are the 
    center of gravity on *x* and *y* axis normalized by width and height,
    respectively. The following seven entries are the 
    *normalized central moments* (*u20,u02,u11,u30,u12,u21,u03*). For their
    definition, see Gonzalez, Woods: \"Digital Image Processing\",
    Prentice Hall, second edition (2002).

    +---------------------------+
    | **Invariant to:**         |  
    +-------+----------+--------+
    | scale | rotation | mirror | 
    +-------+----------+--------+
    |   X   |          |        |
    +-------+----------+--------+
    """
    return_type = FloatVector(length=9)

class nholes(Feature):
    """
    Computes for each row or column the average number of white runs not
	touching the border. From these values, the average over all rows and
	all columns is returned.

    The elements of the returned ``FloatVector`` are:

    0. vertical
    1. horizontal

    +---------------------------+
    | **Invariant to:**         |  
    +-------+----------+--------+
    | scale | rotation | mirror | 
    +-------+----------+--------+
    |   X   |          |   X    |
    +-------+----------+--------+
    """
    return_type = FloatVector(length=2)

class nholes_extended(Feature):
    """
    Divides the image into four strips and then does a nholes_
    analysis on each of those strips. This is first done vertically
    and then horizontally, resulting in a total of eight feature values.

    The elements of the returned ``FloatVector`` are:

    0 - 3
      vertical ``nholes`` for each of the strips in order left to right.
    4 - 7
      horizonal ``nholes`` for each of the strips in order top to bottom.

    +---------------------------+
    | **Invariant to:**         |  
    +-------+----------+--------+
    | scale | rotation | mirror | 
    +-------+----------+--------+
    |   X   |          |        |
    +-------+----------+--------+
    """
    return_type = FloatVector(length=8)

class volume(Feature):
    """
    The percentage of black pixels within the rectangular bounding box
    of the image.  Result in range (0, 1].

    +---------------------------+
    | **Invariant to:**         |  
    +-------+----------+--------+
    | scale | rotation | mirror | 
    +-------+----------+--------+
    |   X   |          |   X    |
    +-------+----------+--------+
    """
    pass

class area(Feature):
    """
    The area of the bounding box (i.e. *nrows* * *ncols*).

    +---------------------------+
    | **Invariant to:**         |  
    +-------+----------+--------+
    | scale | rotation | mirror | 
    +-------+----------+--------+
    |       |          |   X    |
    +-------+----------+--------+
    """
    pass

class aspect_ratio(Feature):
    """
    The aspect ratio of the bounding box (i.e. *ncols* / *nrows*).

    This feature is scale invariant.

    +---------------------------+
    | **Invariant to:**         |  
    +-------+----------+--------+
    | scale | rotation | mirror | 
    +-------+----------+--------+
    |   X   |          |   X    |
    +-------+----------+--------+
    """
    pass
    
class nrows_feature(Feature):
    """
    Simply the number of rows. As this feature is *not* scale
    invariant, it is helpful for distinguishing similar symbols of
    different size.

    +---------------------------+
    | **Invariant to:**         |  
    +-------+----------+--------+
    | scale | rotation | mirror | 
    +-------+----------+--------+
    |       |          |   X    |
    +-------+----------+--------+
    """
    pass

class ncols_feature(Feature):
    """
    Simply the number of cols. As this feature is *not* scale
    invariant, it is helpful for distinguishing similar symbols of
    different size.

    +---------------------------+
    | **Invariant to:**         |  
    +-------+----------+--------+
    | scale | rotation | mirror | 
    +-------+----------+--------+
    |       |          |   X    |
    +-------+----------+--------+
    """
    pass

class compactness(Feature):
    """
    Compactness is the volume to surface ratio. Highly ornate connected
    components have a low compactness, whereas a perfect circle has a 
    high compactness. The present implementation of this feature in
    Gamera does however not return the compactness, but its *inverse*,
    i.e. the surface to volume ratio.

    Since this function requires allocation and deallocation of
    memory, it is relatively slow.  However, it has proven to be a
    very useful feature in many cases.

    This feature is not scale invariant, because, as
    the image is scaled by *a*, the surface increases proportional to *a*
    while the volume grows with *a^2*.
    This is currently not corrected for.

    +---------------------------+
    | **Invariant to:**         |  
    +-------+----------+--------+
    | scale | rotation | mirror | 
    +-------+----------+--------+
    |       |    X     |   X    |
    +-------+----------+--------+
    """
    pass
    
class volume16regions(Feature):
    """
    Divides the image into a 4 x 4 grid of 16 regions and calculates
    the volume within each. This feature is also known as \"zoning\" method.

    +---------------------------+
    | **Invariant to:**         |  
    +-------+----------+--------+
    | scale | rotation | mirror | 
    +-------+----------+--------+
    |   X   |          |        |
    +-------+----------+--------+
    """
    return_type = FloatVector(length=16)

class volume64regions(Feature):
    """
    Divides the image into an 8 x 8 grid of 64 regions and calculates
    the volume within each. This feature is also known as \"zoning\" method.

    +---------------------------+
    | **Invariant to:**         |  
    +-------+----------+--------+
    | scale | rotation | mirror | 
    +-------+----------+--------+
    |   X   |          |        |
    +-------+----------+--------+
    """
    return_type = FloatVector(length=64)

class zernike_moments(Feature):
    """
    Computes the absolute values of the Normalized Zernike Moments up to
    order six.

    Zernike moments are the scalar product in the complex L2
    Hilbert space between the image and the complex Zernike polynomials.
    This results in complex values, the absolute values of which are
    invariant under image rotation and reflection. To make them additionally
    scale invariant, different normalization schemes have been suggested.
    The present implementation normalizes the Zernike moments by
    division with the zeroeth geometric moment *m00*, which results
    in an approximate scale invariance according to
    S. Belkasim, E. Hassan, T. Obeidi: \"Explicit invariance of Cartesian
    Zernike moments.\" Pattern Recognition Letters 28, pp. 1969-1980 (2007)

    The return values are the absolute values of
    *A20, A22, A31, A33, A40, A42, A44, A51, A53, A54, A60, A62, A64, A66*.
    The moments *A00* and *A11* are not computed because these are constant
    under the used normalization scheme.

    +---------------------------+
    | **Invariant to:**         |  
    +-------+----------+--------+
    | scale | rotation | mirror | 
    +-------+----------+--------+
    |   X   |    X     |   X    |
    +-------+----------+--------+
    """
    author = "Robert Butz, Fabian Schmitt, Christoph Dalitz"
    return_type = FloatVector(length=14)

class zernike_moments_plugin(PluginFunction):
    """
    Computes the absolute values of the Normalized Zernike Moments up to
    the given order. This is separate from the feature *zernike_moments*,
    because features only support onebit images, so we need a different
    function for greyscale images.

    Zernike moments are the scalar product in the complex L2
    Hilbert space between the image and the complex Zernike polynomials.
    This results in complex values, the absolute values of which are
    invariant under image rotation and reflection. To make them additionally
    scale invariant, different normalization schemes have been suggested.
    The present implementation normalizes the Zernike moments by
    division with the zeroeth geometric moment *m00*, which results
    in an approximate scale invariance according to
    S. Belkasim, E. Hassan, T. Obeidi: \"Explicit invariance of Cartesian
    Zernike moments.\" Pattern Recognition Letters 28, pp. 1969-1980 (2007)

    The return values are the absolute values of
    *A20, A22, A31, A33, A40, A42, A44, A51, A53, A54, A60, A62, A64, A66*.
    The moments *A00* and *A11* are not computed because these are constant
    under the used normalization scheme.
    """
    category="Features"
    author = "Christoph Dalitz, Robert Butz, Fabian Schmitt"
    self_type = ImageType([GREYSCALE])
    args = Args([Int("order", default=6)])
    return_type = FloatVector("zernike_moments")
    def __call__(image, order = 6):
        return _features.zernike_moments_plugin(image, order)
    __call__ = staticmethod(__call__)

class skeleton_features(Feature):
    """
    Generates a number of features based on the skeleton of an image.
    First, the image in skeletonized using the Lee and Chen algorithm,
    which guarantees that the pixels of the resulting skeleton are
    never more than 4-connected.  Then, this skeleton is analysed for
    a number of properties:
    
    0. Number of X joints (4-connected pixels)
    1. Number of T joints (3-connected pixels)
    2. Average number of bend points (pixels which do not form a horizontal or
       vertical line with its neighbors)
    3. Number of end points (1-connected pixels)
    4. Number of *x*-axis crossings with respect to the *x*-axis through the center of mass
    5. Number of *y*-axis crossings with respect to the *y*-axis through the center of mass

    +---------------------------+
    | **Invariant to:**         |  
    +-------+----------+--------+
    | scale | rotation | mirror | 
    +-------+----------+--------+
    |   X   |          |   X    |
    +-------+----------+--------+
    """
    return_type = FloatVector(length=6)

class top_bottom(Feature):
    """
    Features useful only for segmentation-free analysis.  Currently,
    the first feature is the first row containing a black pixel, and
    the second feature is the last row containing a black pixel.

    +---------------------------+
    | **Invariant to:**         |  
    +-------+----------+--------+
    | scale | rotation | mirror | 
    +-------+----------+--------+
    |   X   |          |   X    |
    +-------+----------+--------+
    """
    return_type = FloatVector(length=2)

class diagonal_projection(Feature):
    """
    Computes the ratio of sum of the interquartile (center plusminus half width)
    rotated projections at +/- 45 degrees.

    +---------------------------+
    | **Invariant to:**         |  
    +-------+----------+--------+
    | scale | rotation | mirror | 
    +-------+----------+--------+
    |   X   |          |        |
    +-------+----------+--------+
    """
    return_type = FloatVector(length=1)
    author = "Andrew Hankinson"


class generate_features(PluginFunction):
    """
    Generates features for the image by calling a number of feature
    functions and storing the results in the image's ``features``
    member variable (a Python ``array``).
    
    *features*
      Optional.  A list of feature function names.  If not given, the
      previously set feature functions will be used.  If none were
      previously given, all available feature functions will be used.
      Using all feature functions can also be forced by passing
      ``'all'``.

    .. warning:: For efficiency, if the given feature functions match
       those that have been already generated for the image, the
       features are *not* recalculated.  If you want to force
       recalculation, pass the optional argument ``force=True``.
    """
    category = "Utility"
    pure_python = True
    self_type = ImageType([ONEBIT])
    args = Args([Class('features', list), Check('force')])
    return_type = None
    cache = {}
    def __call__(self, features=None, force=False):
      if features is None:
         features = self.get_feature_functions()
      if self.feature_functions == features and not force:
         return
      self.feature_functions = features
      features, num_features = features
      if len(self.features) != num_features:
          if not generate_features.cache.has_key(num_features):
              generate_features.cache[num_features] = [0] * num_features
          self.features = array.array('d', generate_features.cache[num_features])
      offset = 0
      for name, function in features:
          function.__call__(self, offset)
          offset += function.return_type.length
    __call__ = staticmethod(__call__)

class FeaturesModule(PluginModule):
    category = "Features"
    cpp_headers=["features.hpp"]
    functions = [black_area, moments, nholes,
                 nholes_extended, volume, area,
                 aspect_ratio, nrows_feature, ncols_feature, compactness,
                 volume16regions, volume64regions,
                 generate_features, zernike_moments, zernike_moments_plugin,
                 skeleton_features, top_bottom, diagonal_projection]
    author = "Michael Droettboom and Karl MacMillan"
    url = "http://gamera.sourceforge.net/"
module = FeaturesModule()

def get_features_length(features):
    """
    Given a list of feature functions return the number of features
    that will be generated. This function is necessary because each
    features 'function' can return multiple individual float values.
    """
    from gamera import core
    ff = core.ImageBase.get_feature_functions(features)
    return ff[1]

def generate_features_list(list, features='all'):
   """
   Generate features on a list of images.

   *features*
     Follows the same rules as for generate_features_.
   """
   from gamera import core, util
   ff = core.Image.get_feature_functions(features)
   progress = util.ProgressFactory("Generating features...", len(list) / 10)
   try:
      for i, glyph in enumerate(list):
         glyph.generate_features(ff)
         if i % 10 == 0:
             progress.step()
   finally:
       progress.kill()

generate_features = generate_features()

del Feature



# Copyright (c) 1999-2007 Gary Strangman; All Rights Reserved.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
#
# Comments and/or additions are welcome (send e-mail to:
# strang@nmr.mgh.harvard.edu).
#
"""
Defines a number of functions for pseudo-command-line OS functionality.

    cd(directory)
    pwd                 <-- can be used WITHOUT parens
    ls(d='.')
    rename(from,to)
    get(namepatterns,verbose=1)
    getstrings(namepatterns,verbose=1)
    put(outlist,filename,writetype='w')
    aget(namepatterns,verbose=1)
    aput(outarray,filename,writetype='w')
    bget(filename,numslices=1,xsize=64,ysize=64)
    braw(filename,btype)
    bput(outarray,filename,writeheader=0,packstring='h',writetype='wb')
    mrget(filename)
    find_dirs(sourcedir)
"""

## CHANGES:
## =======
## 08-07-23 ... added getcsv() to more easily handle tabbed files
## 07-11-26 ... more numpy conversion work
## 06-08-07 ... converted to numpy, changed version to 0.6
## 06-02-03 ... added add2afnihistory() to load modify afni HISTORY_NOTEs,
##              and added that option to array2afni output
## 04-06-14 ... added getafniparam() to load in afni values from HEAD files
## 03-04-09 ... fixed brikget to load based on datatype, changed array2afni
##              so that the default sliceorder is altplus, not seqplus
## 02-11-20 ... added binget(), binput(), array2afni(), version 0.5
## 02-10-20 ... added find_dirs() function, changed version to 0.4
## 01-11-15 ... changed aput() and put() to accept a delimiter
## 01-04-19 ... added oneperline option to put() function
## 99-11-07 ... added DAs quick flat-text-file loaders, load() and fload()
## 99-11-01 ... added version number (0.1) for distribution
## 99-08-30 ... Put quickload in here
## 99-06-27 ... Changed bget thing back ... confused ...
## 99-06-24 ... exchanged xsize and ysize in bget for non-square images (NT??)
##              modified bget to raise an IOError when file not found
## 99-06-12 ... added load() and save() aliases for aget() and aput() (resp.)
## 99-04-13 ... changed aget() to ignore (!!!!) lines beginning with # or %
## 99-01-17 ... changed get() so ints come in as ints (not floats)
##



try:
    import mmapfile
except Exception:
    pass

import pstat
import glob, re, string, types, os, struct, copy, time, tempfile, sys
import csv
from types import *
import numpy as np

__version__ = 0.7

def wrap(f):
    """
Wraps a function so that if it's entered *by itself*
in the interpreter without ()'s, it gets called anyway
"""
    class W:
        def __init__(self, f):
            self.f = f
        def __repr__(self):
            x =apply(self.f)
            if x:
                return repr(x)
            else:
                return ''
    return W(f)


def cd (directory):
    """
Changes the working python directory for the interpreter.

Usage:   cd(directory)      where 'directory' is a string
"""
    os.chdir(directory)
    return


def pwd():
    """
Changes the working python directory for the interpreter.

Usage:   pwd       (no parens needed)
"""
    return os.getcwd()
pwd = wrap(pwd)


def ls(d='.'):
    """
Produces a directory listing.  Default is the current directory.

Usage:   ls(d='.')
"""
    os.system('ls '+d)
    return None


def rename(source, dest):
    """
Renames files specified by UNIX inpattern to those specified by UNIX
outpattern.  Can only handle a single '*' in the two patterns!!!

Usage:   rename (source, dest)     e.g., rename('*.txt', '*.c')
"""
    infiles = glob.glob(source)
    outfiles = []
    incutindex = string.index(source,'*')
    outcutindex = string.index(source,'*')
    findpattern1 = source[0:incutindex]
    findpattern2 = source[incutindex+1:]
    replpattern1 = dest[0:incutindex]
    replpattern2 = dest[incutindex+1:]
    for fname in infiles:
        if incutindex > 0:
            newname = re.sub(findpattern1,replpattern1,fname,1)
        if outcutindex < len(dest)-1:
            if incutindex > 0:
                lastone = string.rfind(newname,replpattern2)
                newname = newname[0:lastone] + re.sub(findpattern2,replpattern2,fname[lastone:],1)
            else:
                lastone = string.rfind(fname,findpattern2)
                if lastone <> -1:
                    newname = fname[0:lastone]
                    newname = newname + re.sub(findpattern2,replpattern2,fname[lastone:],1)
        print fname, newname
        os.rename(fname,newname)
    return


def get (namepatterns,verbose=1):
    """
Loads a list of lists from text files (specified by a UNIX-style
wildcard filename pattern) and converts all numeric values to floats.
Uses the glob module for filename pattern conversion.  Loaded filename
is printed if verbose=1.

Usage:   get (namepatterns,verbose=1)
Returns: a 1D or 2D list of lists from whitespace delimited text files
         specified by namepatterns; numbers that can be converted to floats
         are so converted
"""
    fnames = []
    if type(namepatterns) in [ListType,TupleType]:
        for item in namepatterns:
            fnames = fnames + glob.glob(item)
    else:
        fnames = glob.glob(namepatterns)

    if len(fnames) == 0:
        if verbose:
            print 'NO FILENAMES MATCH ('+namepatterns+') !!'
        return None

    if verbose:
        print fnames             # so user knows what has been loaded
    elements = []
    for i in range(len(fnames)):
        file = open(fnames[i])
        newelements = map(string.split,file.readlines())
        for i in range(len(newelements)):
            for j in range(len(newelements[i])):
                try:
                    newelements[i][j] = string.atoi(newelements[i][j])
                except ValueError:
                    try:
                        newelements[i][j] = string.atof(newelements[i][j])
                    except Exception:
                        pass
        elements = elements + newelements
    if len(elements)==1:  elements = elements[0]
    return elements


def getstrings (namepattern,verbose=1):
    """
Loads a (set of) text file(s), with all elements left as string type.
Uses UNIX-style wildcards (i.e., function uses glob).  Loaded filename
is printed if verbose=1.

Usage:   getstrings (namepattern, verbose=1)
Returns: a list of strings, one per line in each text file specified by
         namepattern
"""
    fnames = glob.glob(namepattern)
    if len(fnames) == 0:
        if verbose:
            print 'NO FILENAMES MATCH ('+namepattern+') !!'
        return None
    if verbose:
        print fnames
    elements = []
    for filename in fnames:
        file = open(filename)
        newelements = map(string.split,file.readlines())
        elements = elements + newelements
    return elements


def getcsv (namepatterns, delimiter='\t', verbose=1):
    """
Loads a list of lists from text files (specified by a UNIX-style
wildcard filename pattern) and converts all numeric values to floats.
Uses the glob module for filename pattern conversion.  Loaded filename
is printed if verbose=1.

Usage:   get (namepatterns, delimiter='\t', verbose=1)
Returns: a 1D or 2D list of lists from whitespace delimited text files
         specified by namepatterns; numbers that can be converted to floats
         are so converted
"""
    fnames = []
    if type(namepatterns) in [ListType,TupleType]:
        for item in namepatterns:
            fnames = fnames + glob.glob(item)
    else:
        fnames = glob.glob(namepatterns)

    if len(fnames) == 0:
        if verbose:
            print 'NO FILENAMES MATCH ('+namepatterns+') !!'
        return None

    if verbose:
        print fnames             # so user knows what has been loaded
    elements = []
    for i in range(len(fnames)):
        file = csv.reader(open(fnames[i]),delimiter=delimiter)
        newelements = [row for row in file]
        for i in range(len(newelements)):
            for j in range(len(newelements[i])):
                try:
                    newelements[i][j] = string.atoi(newelements[i][j])
                except Exception:
                    try:
                        newelements[i][j] = string.atof(newelements[i][j])
                    except Exception:
                        pass
        elements = elements + newelements
    if len(elements)==1:  elements = elements[0]
    return elements


def getrec(namepattern,verbose=0):
    """
Loads a numpy.recarray from a text file with the first row as the col names.


Usage:   getrec(namepatterns,verbose=0)
Returns: a 2D recarray
"""
    fname = glob.glob(namepattern)

    if len(fname) == 0:
        if verbose:
            print 'NO FILENAMES MATCH ('+namepatterns+') !!'
        return None

    if verbose:
        print fnames        # so user knows what has been loaded

    d = get(fname[0])
    h = d[0]                # ASSUME FIRST ROW HOLDS COLUMN NAMES
    d = d[1:]

    return np.rec.fromrecords(d,names=h)


def put (outlist,fname,writetype='w',oneperline=0,delimiter=' '):
    """
Writes a passed mixed-type list (str and/or numbers) to an output
file, and then closes the file.  Default is overwrite the destination
file.

Usage:   put (outlist,fname,writetype='w',oneperline=0,delimiter=' ')
Returns: None
"""
    if type(outlist) in [np.ndarray]:
        aput(outlist,fname,writetype)
        return
    if type(outlist[0]) not in [ListType,TupleType]:  # 1D list
        outfile = open(fname,writetype)
        if not oneperline:
            outlist = pstat.list2string(outlist,delimiter)
            outfile.write(outlist)
            outfile.write('\n')
        else:  # they want one element from the list on each file line
            for item in outlist:
                outfile.write(str(item)+'\n')
        outfile.close()
    else:                                             # 2D list (list-of-lists)
        outfile = open(fname,writetype)
        for row in outlist:
            outfile.write(pstat.list2string(row,delimiter))
            outfile.write('\n')
        outfile.close()
    return None


def isstring(x):
    if type(x)==StringType:
        return 1
    else:
        return 0



def aget (namepattern,verbose=1):
    """
Loads an array from 2D text files (specified by a UNIX-style wildcard
filename pattern).  ONLY 'GET' FILES WITH EQUAL NUMBERS OF COLUMNS
ON EVERY ROW (otherwise returned array will be zero-dimensional).

Usage:   aget (namepattern)
Returns: an array of integers, floats or objects (type='O'), depending on the
         contents of the files specified by namepattern
"""
    fnames = glob.glob(namepattern)
    if len(fnames) == 0:
        if verbose:
            print 'NO FILENAMES MATCH ('+namepattern+') !!'
            return None
    if verbose:
        print fnames
    elements = []
    for filename in fnames:
        file = open(filename)
        newelements = file.readlines()
        del_list = []
        for row in range(len(newelements)):
            if (newelements[row][0]=='%' or newelements[row][0]=='#'
                or len(newelements[row])==1 or newelements[row][0]=='\r'):
                del_list.append(row)
        del_list.reverse()
        for i in del_list:
            newelements.pop(i)
        newelements = map(string.split,newelements)
        for i in range(len(newelements)):
            for j in range(len(newelements[i])):
                try:
                    newelements[i][j] = string.atof(newelements[i][j])
                except Exception:
                    pass
        elements = elements + newelements
    for row in range(len(elements)):
        if np.add.reduce(np.array(map(isstring,elements[row])))==len(elements[row]):
            print "A row of strings was found.  Returning a LIST."
            return elements
    try:
        elements = np.array(elements)
    except TypeError:
        elements = np.array(elements,dtype='O')
    return elements


def aput (outarray,fname,writetype='w',delimit=' '):
    """
Sends passed 1D or 2D array to an output file and closes the file.

Usage:   aput (outarray,fname,writetype='w',delimit=' ')
Returns: None
"""
    outfile = open(fname,writetype)
    if len(outarray.shape) == 1:
        outarray = outarray[np.newaxis,:]
    if len(outarray.shape) > 2:
        raise TypeError, "put() and aput() require 1D or 2D arrays.  Otherwise use some kind of pickling."
    else: # must be a 2D array
        for row in outarray:
            outfile.write(string.join(map(str,row),delimit))
            outfile.write('\n')
        outfile.close()
    return None


def bget(imfile,shp=None,unpackstr=np.int16,bytesperpixel=2.0,sliceinit=0):
    """
Reads in a binary file, typically with a .bshort or .bfloat extension.
If so, the last 3 parameters are set appropriately.  If not, the last 3
parameters default to reading .bshort files (2-byte integers in big-endian
binary format).

Usage:   bget(imfile,shp=None,unpackstr=np.int16,bytesperpixel=2.0,sliceinit=0)
"""
    if imfile[:3] == 'COR':
        return CORget(imfile)
    if imfile[-2:] == 'MR':
        return mrget(imfile,unpackstr)
    if imfile[-4:] == 'BRIK':
        return brikget(imfile,unpackstr,shp)
    if imfile[-3:] in ['mnc','MNC','inc','INC']:
        return mincget(imfile,unpackstr,shp)
    if imfile[-3:] == 'img':
        return mghbget(imfile,unpackstr,shp)
    if imfile[-6:] == 'bshort' or imfile[-6:] == 'bfloat':
        if shp == None:
            return mghbget(imfile,unpackstr=unpackstr,bytesperpixel=bytesperpixel,sliceinit=sliceinit)
        else:
            return mghbget(imfile,shp[0],shp[1],shp[2],unpackstr,bytesperpixel,sliceinit)


def CORget(infile):
    """
Reads a binary COR-nnn file (flattening file).

Usage:   CORget(imfile)
Returns: 2D array of 16-bit ints
"""
    d=braw(infile,np.int8)
    d.shape = (256,256)
    d = np.where(d>=0,d,256+d)
    return d


def mincget(imfile,unpackstr=np.int16,shp=None):
    """
Loads in a .MNC file.

Usage:  mincget(imfile,unpackstr=np.int16,shp=None)  default shp = -1,20,64,64
"""
    if shp == None:
        shp = (-1,20,64,64)
    os.system('mincextract -short -range 0 4095 -image_range 0 4095 ' +
              imfile+' > minctemp.bshort')
    try:
        d = braw('minctemp.bshort',unpackstr)
    except Exception:
        print "Couldn't find file:  "+imfile
        raise IOError, "Couldn't find file in mincget()"

    print shp, d.shape
    d.shape = shp
    os.system('rm minctemp.bshort')
    return d


def brikget(imfile,unpackstr=np.int16,shp=None):
    """
Gets an AFNI BRIK file.

Usage:  brikget(imfile,unpackstr=np.int16,shp=None)  default shp: (-1,48,61,51)
"""
    if shp == None:
        shp = (-1,48,61,51)
    try:
        file = open(imfile, "rb")
    except Exception:
        print "Couldn't find file:  "+imfile
        raise IOError, "Couldn't find file in brikget()"
    try:
        header = imfile[0:-4]+'HEAD'
        lines = open(header).readlines()
        for i in range(len(lines)):
            if string.find(lines[i],'DATASET_DIMENSIONS') <> -1:
                dims = string.split(lines[i+2][0:string.find(lines[i+2],' 0')])
                dims = map(string.atoi,dims)
            if string.find(lines[i],'BRICK_FLOAT_FACS') <> -1:
                count = string.atoi(string.split(lines[i+1])[2])
                mults = []
                for j in range(int(np.ceil(count/5.))):
                    mults += map(string.atof,string.split(lines[i+2+j]))
                mults = np.array(mults)
            if string.find(lines[i],'BRICK_TYPES') <> -1:
                first5 = lines[i+2]
                first5 = map(string.atoi,string.split(first5))
                if first5[0] == 0:
                    unpackstr = np.uint8
                elif first5[0] == 1:
                    unpackstr = np.int16
                elif first5[0] == 3:
                    unpackstr = np.float32
                elif first5[0] == 5:
                    unpackstr = np.complex32
        dims.reverse()
        shp = [-1]+dims
    except IOError:
        print "No header file.  Continuing ..."
    lines = None

    print shp
    print 'Using unpackstr:',unpackstr  #,', bytesperpixel=',bytesperpixel

    file = open(imfile, "rb")
    bdata = file.read()

    # the > forces big-endian (for or from Sun/SGI)
    bdata = np.fromstring(bdata,unpackstr)
#    littleEndian = ( struct.pack('i',1)==struct.pack('<i',1) )
    if (max(bdata)>1e30):
        bdata = bdata.byteswap()
    try:
        bdata.shape = shp
    except Exception:
        print 'Incorrect shape ...',shp,len(bdata)
        raise ValueError, 'Incorrect shape for file size'
    if len(bdata) == 1:
        bdata = bdata[0]

    if np.sum(mults) == 0:
        return bdata
    try:
        multshape = [1]*len(bdata.shape)
        for i in range(len(bdata.shape)):
            if len(mults) == bdata.shape[i]:
                multshape[i] = len(mults)
                break
        mults.shape = multshape
        return bdata*mults
    except Exception:
        return bdata

def mghbget(imfile,numslices=-1,xsize=64,ysize=64,
           unpackstr=np.int16,bytesperpixel=2.0,sliceinit=0):
    """
Reads in a binary file, typically with a .bshort or .bfloat extensionp.
If so, the last 3 parameters are set appropriately.  If not, the last 3
parameters default to reading .bshort files (2-byte integers in big-endian
binary format).

Usage:   mghbget(imfile, numslices=-1, xsize=64, ysize=64,
                unpackstr=np.int16, bytesperpixel=2.0, sliceinit=0)
"""
    try:
        file = open(imfile, "rb")
    except Exception:
        print "Couldn't find file:  "+imfile
        raise IOError, "Couldn't find file in bget()"
    try:
        header = imfile[0:-6]+'hdr'
        vals = get(header,0)  # '0' means no missing-file warning msg
        if type(vals[0]) == ListType:  # it's an extended header
            xsize = int(vals[0][0])
            ysize = int(vals[0][1])
            numslices = int(vals[0][2])
        else:
            xsize = int(vals[0])
            ysize = int(vals[1])
            numslices = int(vals[2])
    except Exception:
        print "No header file.  Continuing ..."

    suffix = imfile[-6:]
    if suffix == 'bshort':
        pass
    elif suffix[-3:] == 'img':
        pass
    elif suffix == 'bfloat':
        unpackstr = np.float32
        bytesperpixel = 4.0
        sliceinit = 0.0
    else:
        print 'Not a bshort, bfloat or img file.'
        print 'Using unpackstr:',unpackstr,', bytesperpixel=',bytesperpixel

    imsize = xsize*ysize
    file = open(imfile, "rb")
    bdata = file.read()

    numpixels = len(bdata) / bytesperpixel
    if numpixels%1 != 0:
        raise ValueError, "Incorrect file size in fmri.bget()"
    else:  # the > forces big-endian (for or from Sun/SGI)
        bdata = np.fromstring(bdata,unpackstr)
#        littleEndian = ( struct.pack('i',1)==struct.pack('<i',1) )
#        if littleEndian:
#            bdata = bdata.byteswap()
        if (max(bdata)>1e30):
            bdata = bdata.byteswap()
    if suffix[-3:] == 'img':
        if numslices == -1:
            numslices = len(bdata)/8200  # 8200=(64*64*2)+8 bytes per image
            xsize = 64
            ysize = 128
        slices = np.zeros((numslices,xsize,ysize),np.int32)
        for i in range(numslices):
            istart = i*8 + i*xsize*ysize
            iend = i*8 + (i+1)*xsize*ysize
            print i, istart,iend
            slices[i] = np.reshape(np.array(bdata[istart:iend]),(xsize,ysize))
    else:
        if numslices == 1:
            slices = np.reshape(np.array(bdata),[xsize,ysize])
        else:
            slices = np.reshape(np.array(bdata),[numslices,xsize,ysize])
    if len(slices) == 1:
        slices = slices[0]
    return slices


def braw(fname,btype,shp=None):
    """
Opens a binary file, unpacks it, and returns a flat array of the
type specified.  Use Numeric types ... np.float32, np.int64, etc.

Usage:   braw(fname,btype,shp=None)
Returns: flat array of floats, or ints (if btype=np.int16)
"""
    file = open(fname,'rb')
    bdata = file.read()
    bdata = np.fromstring(bdata,btype)
#    littleEndian = ( struct.pack('i',1)==struct.pack('<i',1) )
#    if littleEndian:
#        bdata = bdata.byteswap()  # didn't used to need this with '>' above
    if (max(bdata)>1e30):
        bdata = bdata.byteswap()
    if shp:
        try:
            bdata.shape = shp
            return bdata
        except Exception:
            pass
    return np.array(bdata)


def glget(fname,btype):
    """
Load in a file containing pixels from glReadPixels dump.

Usage:   glget(fname,btype)
Returns: array of 'btype elements with shape 'shape', suitable for im.ashow()
"""
    d = braw(fname,btype)
    d = d[8:]
    f = open(fname,'rb')
    shp = f.read(8)
    f.close()
    shp = np.fromstring(shp,np.int32)
    shp[0],shp[1] = shp[1],shp[0]
    try:
        carray = np.reshape(d,shp)
        return
    except Exception:
        pass
    try:
        r = d[0::3]+0
        g = d[1::3]+0
        b = d[2::3]+0
        r.shape = shp
        g.shape = shp
        b.shape = shp
        carray = np.array([r,g,b])
    except Exception:
        outstr = "glget: shape not correct for data of length "+str(len(d))
        raise ValueError, outstr
    return carray


def mget(fname,btype):
    """
Load in a file that was saved from matlab

Usage:   mget(fname,btype)
"""
    d = braw(fname,btype)
    try:
        header = fname[0:-6]+'hdr'
        vals = get(header,0)  # '0' means no missing-file warning msg
        if type(vals[0]) == ListType:  # it's an extended header
            xsize = int(vals[0][0])
            ysize = int(vals[0][1])
            numslices = int(vals[0][2])
        else:
            xsize = int(vals[0])
            ysize = int(vals[1])
            numslices = int(vals[2])
        print xsize,ysize,numslices, d.shape
    except Exception:
        print "No header file.  Continuing ..."
    if numslices == 1:
        d.shape = [ysize,xsize]
        return np.transpose(d)*1
    else:
        d.shape = [numslices,ysize,xsize]
        return np.transpose(d)*1


def mput(outarray,fname,writeheader=0,btype=np.int16):
    """
Save a file for use in matlab.
"""
    outarray = np.transpose(outarray)
    outdata = np.ravel(outarray).astype(btype)
    outdata = outdata.tostring()
    outfile = open(fname,'wb')
    outfile.write(outdata)
    outfile.close()
    if writeheader == 1:
        try:
            suffixindex = string.rfind(fname,'.')
            hdrname = fname[0:suffixindex]
        except ValueError:
            hdrname = fname
        if len(outarray.shape) == 2:
            hdr = [outarray.shape[1],outarray.shape[0], 1, 0]
        else:
            hdr = [outarray.shape[2],outarray.shape[1],outarray.shape[0], 0,'\n']
        print hdrname+'.hdr'
        outfile = open(hdrname+'.hdr','w')
        outfile.write(pstat.list2string(hdr))
        outfile.close()
    return None


def bput(outarray,fname,writeheader=0,packtype=np.int16,writetype='wb'):
    """
Writes the passed array to a binary output file, and then closes
the file.  Default is overwrite the destination file.

Usage:   bput (outarray,filename,writeheader=0,packtype=np.int16,writetype='wb')
"""
    suffix = fname[-6:]
    if suffix == 'bshort':
        packtype = np.int16
    elif suffix == 'bfloat':
        packtype = np.float32
    else:
        print 'Not a bshort or bfloat file.  Using packtype=',packtype

    outdata = np.ravel(outarray).astype(packtype)
#    littleEndian = ( struct.pack('i',1)==struct.pack('<i',1) )
#    if littleEndian:
#        outdata = outdata.byteswap()
    outdata = outdata.tostring()
    outfile = open(fname,writetype)
    outfile.write(outdata)
    outfile.close()
    if writeheader == 1:
        try:
            suffixindex = string.rfind(fname,'.')
            hdrname = fname[0:suffixindex]
        except ValueError:
            hdrname = fname
        if len(outarray.shape) == 2:
            hdr = [outarray.shape[0],outarray.shape[1], 1, 0]
        else:
            hdr = [outarray.shape[1],outarray.shape[2],outarray.shape[0], 0,'\n']
        print hdrname+'.hdr'
        outfile = open(hdrname+'.hdr','w')
        outfile.write(pstat.list2string(hdr))
        outfile.close()
    return None


def mrget(fname,datatype=np.int16):
    """
Opens a binary .MR file and clips off the tail data portion of it, returning
the result as an array.

Usage:   mrget(fname,datatype=np.int16)
"""
    d = braw(fname,datatype)
    if len(d) > 512*512:
        return np.reshape(d[-512*512:],(512,512))
    elif len(d) > 320*320:
        return np.reshape(d[-320*320:],(320,320))
    elif len(d) > 256*256:
        return np.reshape(d[-256*256:],(256,256))
    elif len(d) > 128*128:
        return np.reshape(d[-128*128:],(128,128))
    elif len(d) > 64*64:
        return np.reshape(d[-64*64:],(64,64))
    else:
        return np.reshape(d[-32*32:],(32,32))


def quickload(fname,linestocut=4):
    """
Quickly loads in a long text file, chopping off first n 'linestocut'.

Usage:   quickload(fname,linestocut=4)
Returns: array filled with data in fname
"""
    f = open(fname,'r')
    d = f.readlines()
    f.close()
    print fname,'read inp.'
    d = d[linestocut:]
    d = map(string.split,d)
    print 'Done with string.split on lines.'
    for i in range(len(d)):
        d[i] = map(string.atoi,d[i])
    print 'Conversion to ints done.'
    return np.array(d)

def writedelimited (listoflists, delimiter, file, writetype='w'):
    """
Writes a list of lists in columns, separated by character(s) delimiter
to specified file.  File-overwrite is the default.

Usage:   writedelimited (listoflists,delimiter,filename,writetype='w')
Returns: None
"""
    if type(listoflists[0]) not in [ListType,TupleType]:
        listoflists = [listoflists]
    outfile = open(file,writetype)
    rowstokill = []
    list2print = copy.deepcopy(listoflists)
    for i in range(len(listoflists)):
        if listoflists[i] == ['\n'] or listoflists[i]=='\n' or listoflists[i]=='dashes':
            rowstokill = rowstokill + [i]
    rowstokill.reverse()
    for row in rowstokill:
        del list2print[row]
    maxsize = [0]*len(list2print[0])
    for row in listoflists:
        if row == ['\n'] or row == '\n':
            outfile.write('\n')
        elif row == ['dashes'] or row == 'dashes':
            dashes = [0]*len(maxsize)
            for j in range(len(maxsize)):
                dashes[j] = '------'
            outfile.write(pstat.linedelimited(dashes,delimiter))
        else:
            outfile.write(pstat.linedelimited(row,delimiter))
        outfile.write('\n')
    outfile.close()
    return None

def writecc (listoflists,file,writetype='w',extra=2):
    """
Writes a list of lists to a file in columns, customized by the max
size of items within the columns (max size of items in col, +2 characters)
to specified file.  File-overwrite is the default.

Usage:   writecc (listoflists,file,writetype='w',extra=2)
Returns: None
"""
    if type(listoflists[0]) not in [ListType,TupleType]:
        listoflists = [listoflists]
    outfile = open(file,writetype)
    rowstokill = []
    list2print = copy.deepcopy(listoflists)
    for i in range(len(listoflists)):
        if listoflists[i] == ['\n'] or listoflists[i]=='\n' or listoflists[i]=='dashes':
            rowstokill = rowstokill + [i]
    rowstokill.reverse()
    for row in rowstokill:
        del list2print[row]
    maxsize = [0]*len(list2print[0])
    for col in range(len(list2print[0])):
        items = pstat.colex(list2print,col)
        items = map(pstat.makestr,items)
        maxsize[col] = max(map(len,items)) + extra
    for row in listoflists:
        if row == ['\n'] or row == '\n':
            outfile.write('\n')
        elif row == ['dashes'] or row == 'dashes':
            dashes = [0]*len(maxsize)
            for j in range(len(maxsize)):
                dashes[j] = '-'*(maxsize[j]-2)
            outfile.write(pstat.lineincustcols(dashes,maxsize))
        else:
            outfile.write(pstat.lineincustcols(row,maxsize))
        outfile.write('\n')
    outfile.close()
    return None


def writefc (listoflists,colsize,file,writetype='w'):
    """
Writes a list of lists to a file in columns of fixed size.  File-overwrite
is the default.

Usage:   writefc (listoflists,colsize,file,writetype='w')
Returns: None
"""
    if type(listoflists) == np.ndarray:
        listoflists = listoflists.tolist()
    if type(listoflists[0]) not in [ListType,TupleType]:
        listoflists = [listoflists]
    outfile = open(file,writetype)
    rowstokill = []
    list2print = copy.deepcopy(listoflists)
    for i in range(len(listoflists)):
        if listoflists[i] == ['\n'] or listoflists[i]=='\n' or listoflists[i]=='dashes':
            rowstokill = rowstokill + [i]
    rowstokill.reverse()
    for row in rowstokill:
        del list2print[row]
    n = [0]*len(list2print[0])
    for row in listoflists:
        if row == ['\n'] or row == '\n':
            outfile.write('\n')
        elif row == ['dashes'] or row == 'dashes':
            dashes = [0]*colsize
            for j in range(len(n)):
                dashes[j] = '-'*(colsize)
            outfile.write(pstat.lineincols(dashes,colsize))
        else:
            outfile.write(pstat.lineincols(row,colsize))
        outfile.write('\n')
    outfile.close()
    return None


def load(fname,lines_to_ignore=4,type='i'):
    """
Load in huge, flat, 2D text files.  Can handle differing line-lengths AND
can strip #/% on UNIX (or with a better NT grep).  Requires wc, grep, and
mmapfile.lib/.pyd. Type can be 'i', 'f' or 'd', for ints, floats or doubles,
respectively.  Lines_to_ignore determines how many lines at the start of the
file to ignore (required for non-working grep).

Usage:   load(fname,lines_to_ignore=4,type='i')
Returns: numpy array of specified type
"""
    start = time.time()      ## START TIMER
    if type == 'i':
        intype = int
    elif type in ['f','d']:
        intype = float
    else:
        raise ValueError, "type can be 'i', 'f' or 'd' in load()"

    ## STRIP OUT % AND # LINES
    tmpname = tempfile.NamedTemporaryFile(delete=False).name
    if sys.platform == 'win32':
        # NT VERSION OF GREP DOESN'T DO THE STRIPPING ... SIGH
        cmd = "grep.exe -v \'%\' "+fname+" > "+tmpname
        print cmd
        os.system(cmd)
    else:
        # UNIX SIDE SHOULD WORK
        cmd = "cat "+fname+" | grep -v \'%\' |grep -v \'#\' > "+tmpname
        print cmd
        os.system(cmd)

    ## GET NUMBER OF ROWS, COLUMNS AND LINE-LENGTH, USING WC
    wc = string.split(os.popen("wc "+tmpname).read())
    numlines = int(wc[0]) - lines_to_ignore
    tfp = open(tmpname)
    if lines_to_ignore <> 0:
        for i in range(lines_to_ignore):
            junk = tfp.readline()
    numcols = len(string.split(tfp.readline())) #int(float(wc[1])/numlines)
    tfp.close()

    ## PREPARE INPUT SPACE
    a = np.zeros((numlines*numcols), type)
    block = 65536  # chunk to read, in bytes
    data = mmapfile.mmapfile(tmpname, '', 0)
    if lines_to_ignore <> 0 and sys.platform == 'win32':
        for i in range(lines_to_ignore):
            junk = data.readline()
    i = 0
    d = ' '
    carryover = ''
    while len(d) <> 0:
        d = carryover + data.read(block)
        cutindex = string.rfind(d,'\n')
        carryover = d[cutindex+1:]
        d = d[:cutindex+1]
        d = map(intype,string.split(d))
        a[i:i+len(d)] = d
        i = i + len(d)
    end = time.time()
    print "%d sec" % round(end-start,2)
    data.close()
    os.remove(tmpname)
    return np.reshape(a,[numlines,numcols])


def find_dirs(sourcedir):
    """Finds and returns all directories in sourcedir

Usage:   find_dirs(sourcedir)
Returns: list of directory names (potentially empty)
"""
    files = os.listdir(sourcedir)
    dirs = []
    for fname in files:
        if os.path.isdir(os.path.join(sourcedir,fname)):
            dirs.append(fname)
    return dirs


# ALIASES ...
save = aput



def binget(fname,btype=None):
    """
Loads a binary file from disk. Assumes associated hdr file is in same
locationp. You can force an unpacking type, or else it tries to figure
it out from the filename (4th-to-last character). Hence, readable file
formats are ...

1bin=int8, sbin=int16, ibin=int32, fbin=float32, dbin=float64, etc.

Usage:   binget(fname,btype=None)
Returns: data in file fname of type btype
"""
    file = open(fname,'rb')
    bdata = file.read()
    file.close()

    # if none given, assume character preceeding 'bin' is the unpacktype
    if not btype:
        btype = fname[-4]
    try:
        bdata = np.fromstring(bdata,btype)
    except Exception:
        raise ValueError, "Bad unpacking type."

    # force the data on disk to be LittleEndian (for more efficient PC/Linux use)
    if not np.little_endian:
        bdata = bdata.byteswap()

    try:
        header = fname[:-3]+'hdr'
        vals = get(header,0)  # '0' means no missing-file warning msg
        print vals
        if type(vals[0]) == ListType:  # it's an extended header
            xsize = int(vals[0][0])
            ysize = int(vals[0][1])
            numslices = int(vals[0][2])
        else:
            bdata.shape = vals
    except Exception:
        print "No (or bad) header file. Returning unshaped array."
    return np.array(bdata)



def binput(outarray,fname,packtype=None,writetype='wb'):
    """
Unravels outarray and writes the data to a file, always in LittleEndian
format, along with a header file containing the original data shape. Default
is overwrite the destination file. Tries to figure out packtype from
4th-to-last character in filename. Thus, the routine understands these
file formats ...

1bin=int8, sbin=int16, ibin=int32, fbin=float32, dbin=float64, etc.

Usage:  binput(outarray,filename,packtype=None,writetype='wb')
"""
    if not packtype:
        packtype = fname[-4]

    # a speck of error checking
    if packtype == np.int16 and outarray.dtype.char == 'f':
        # check to see if there's data loss
        if max(np.ravel(outarray)) > 32767 or min(np.ravel(outarray))<-32768:
            print "*** WARNING: CONVERTING FLOAT DATA TO OUT-OF RANGE INT16 DATA"
    outdata = np.ravel(outarray).astype(packtype)

    # force the data on disk to be little_endian (for more efficient PC/Linux use)
    if not np.little_endian:
        outdata = outdata.byteswap()
    outdata = outdata.tostring()
    outfile = open(fname,writetype)
    outfile.write(outdata)
    outfile.close()

    # Now, write the header file
    try:
        suffixindex = string.rfind(fname,'.')
        hdrname = fname[0:suffixindex+2]+'hdr'  # include .s or .f or .1 or whatever
    except ValueError:
        hdrname = fname
    hdr = outarray.shape
    print hdrname
    outfile = open(hdrname,'w')
    outfile.write(pstat.list2string(hdr))
    outfile.close()
    return None

def getafniparam(headfilename,paramname):
    """
Loads in an AFNI header file, and returns the values of 'paramname'.

Usage:   getafniparam(headfile,paramname)
Returns: appropriate "type" for params, or None if fails
"""
    if headfilename[-4:] == 'BRIK':  # if asked for BRIK, change it to HEAD
        headfilename = headfilename[:-4]+'HEAD'
    d = get(headfilename)
    lines = open(headfilename,'r').readlines()
    for i in range(len(lines)):
        if string.find(lines[i],paramname) <> -1:
            count = d[i+1][-1]
            gotten = 0
            result = []
            for j in range(i+2,len(lines)):
                for k in range(len(d[j])):
                    if type(d[j][k]) == StringType:
                        result = d[j][k][1:count]
                        return result
                    else:
                        result.append(d[j][k])
                        gotten += 1
                if gotten == count:
                    break
            return result
    return None


def add2afnihistory(headfilename,newtext):
    """
Adds 'newtext' to HISTORY_NOTE in afni file specified in headfilename.

Usage:   add2afnihistory(headfile,newtext)
Returns: None
"""
    if headfilename[-4:] == 'BRIK':  # if asked for BRIK, change it to HEAD
        headfilename = headfilename[:-4]+'HEAD'
    d = get(headfilename)
    lines = open(headfilename,'r').readlines()
    for i in range(len(lines)):
        if string.find(lines[i],'HISTORY_NOTE') <> -1:
            bytecount = d[i+1][-1]
            oldstr = lines[i+2][:-2]
            date = '[python:***  %s] ' %time.asctime()
            lines[i+2] = oldstr +'\\n' +date +newtext +'~\n'
            lines[i+1] = '  count = %s\n' %str(len(lines[i+2]))
    f = open(headfilename,'w')
    f.writelines(lines)
    f.close()
    return


def array2afni(d,brikprefix,voltype=None,TR=2000,sliceorder='seqplus',geomparent=None,view=None,corrlength=1,briklabels=None,historytext=None):
    """
Converts an array 'd' to an AFNI BRIK/HEAD combo via putbin and to3d. Tries to
guess the AFNI volume type

voltype = {'-anat','-epan','-fim'}
geomparent = filename of the afni BRIK file with the same geometry
view = {'tlrc', 'acpc' or 'orig'}
corrlength = # of images used in the (single-timeseries) correlation (for fico)
briklabels = list of names (strings) to use for brick labels
historytext = string to be appended to the history file, if any

Usage:   array2afni(d,brikprefix,voltype=None,TR=2000,
                    sliceorder='seqplus',geomparent=None,view=None,
                    corrlength=1,briklabels=None,historytext=None)
Returns: None
"""
    # converts numpy typecode()s into appropriate strings for to3d command line
    typecodemapping = {'c':'b',  # character
                       'B':'b',  # UnsignedInt8
                       'f':'f',  # float0, float8, float16, float32
                       'd':'f',  # float64
                       'b':'b',  # int0, int8
                       'h':'',   # int16
                       'i':'i',  # int32
                       'l':'i'}  # int

    # Verify that the data is proper size (3- or 4-D)
    if len(d.shape) not in [3,4]:
        raise ValueError, "A 3D or 4D array is required for array2afni() ... %s" %d.shape

    # Save out the array to a binary file, homebrew style
    if d.dtype.char == np.float64:
        outcode = 'f'
    else:
        outcode = d.dtype.char
    tmpoutname = 'afnitmp.%sbin' % outcode
    binput(d.astype(outcode),tmpoutname)
    if not voltype:
        if len(d.shape) == 3:  # either anatomy or functional
            if d.dtype.char in ['s','i','l']:  # if floats, assume functional
                voltype = '-anat'
            else:
                voltype = '-fim'
        else:  # 4D dataset, must be anatomical timeseries (epan)
            voltype = '-anat'
    if voltype[0] != '-':
        voltype = '-'+voltype
    if len(d.shape) == 3:  # either anatomy or functional
        timepts = 1
        slices = d.shape[0]
        timestr = ''
    elif len(d.shape) == 4:
        if voltype=='-fico':
            timepts = 1
            d = np.reshape(d,[d.shape[0]*d.shape[1],d.shape[2],d.shape[3]])
            slices = d.shape[0]
            timestr = '-statpar %s 1 1 ' % corrlength
        else:
            timepts = d.shape[0]
            slices = d.shape[1]
            timestr = '-time:zt %d %d %0.3f %s ' % (slices,timepts,TR,sliceorder)

    cmd = 'to3d %s -prefix %s -session . ' % (voltype, brikprefix)
    if not view:
        view = 'orig'
    cmd += '-view %s ' % view
    if geomparent:
        cmd += '-geomparent %s ' % geomparent
    cmd += timestr
    cmd += '3D%s:0:0:%d:%d:%d:%s' % (typecodemapping[d.dtype.char],d.shape[-1],d.shape[-2],slices*timepts,tmpoutname)
    print cmd
    os.system(cmd)
    os.remove(tmpoutname)
    os.remove(tmpoutname[:-3]+'hdr')

    if len(d.shape)==4 and briklabels:
        names = ''
        for label in briklabels:
            names += str(label)+'~'
        count = len(names)
        appendstr = """\n\ntype = string-attribute
name = BRICK_LABS
count = %s
'%s""" % (count, names)
        f = open('%s+%s.HEAD' %(brikprefix,view), 'a')
        f.write(appendstr)
        f.close()

        if historytext:
            add2afnihistory('%s+%s.HEAD'%(brikprefix,view),historytext)

def makeDType(exemplar):
    """Return a dtype object based on the given list or dict.
    This is a convenience function -- if you want to do anything sophisticated
    it's best to compose the dtype "by hand".
    """
    if type(exemplar) is dict:
        names = exemplar.keys(); names.sort()
        formats = [np.array(exemplar[key]).dtype for key in names]
        return np.dtype({'names':names, 'formats':formats})
    else:
        formats = ','.join([np.array(val).dtype.str for val in exemplar])
        return np.dtype(formats)



#
#
# Copyright (C) 2001-2005 Ichiro Fujinaga, Michael Droettboom, Karl MacMillan
#               2011      Christoph Dalitz
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

from gamera.plugin import *
import image_utilities, _image_conversion

class to_rgb(PluginFunction):
    """
    Converts the given image to an RGB image according to the following rules:

    - for ONEBIT images, 0 is mapped to (255,255,255) and everything else to (0,0,0)
    - for GREYSCALE and GREY16 images, R=G=B
    - for FLOAT images, the range [min,max] is linearly mapped to the 256 grey values

    Note, converting an image to one of the same type performs a copy operation.
    """
    author = "Michael Droettboom, Karl MacMillan, and Christoph Dalitz"
    self_type = ImageType([ONEBIT, GREYSCALE, FLOAT, GREY16, COMPLEX])
    return_type = ImageType([RGB], "rgb")

    def __call__(self):
        if self.data.pixel_type == RGB:
            return self.image_copy()
        return _image_conversion.to_rgb(self)
    __call__ = staticmethod(__call__)

class to_greyscale(PluginFunction):
    """
    Converts the given image to a GREYSCALE image according to the
    following rules:

    - for ONEBIT images, 0 is mapped to 255 and everything else to 0.
    - for FLOAT images, the range [min,max] is linearly scaled to [0,255]
    - for GREY16 images, the range [0,max] is linearly scaled to [0,255]
    - for RGB images, the luminance is used, which is defined in VIGRA as 0.3*R + 0.59*G + 0.11*B

    Converting an image to one of the same type performs a copy operation.
    """
    author = "Michael Droettboom, Karl MacMillan, and Christoph Dalitz"
    self_type = ImageType([ONEBIT, FLOAT, GREY16, RGB, COMPLEX])
    return_type = ImageType([GREYSCALE], "greyscale")
    doc_examples = [(RGB,)]

    def __call__(self):
        if self.data.pixel_type == GREYSCALE:
            return self.image_copy()
        return _image_conversion.to_greyscale(self)
    __call__ = staticmethod(__call__)

class to_grey16(PluginFunction):
    """
    Converts the given image to a GREY16 image according to the
    following rules:

    - for ONEBIT images, 0 is mapped to 65535 and everything else to 0.
    - for FLOAT images, the range [min,max] is linearly scaled to [0,65535]
    - for GREYSCALE images, pixel values are copied unchanged
    - for RGB images, the luminance is used, which is defined in VIGRA as 0.3*R + 0.59*G + 0.11*B. This results only in a value range [0,255]

    Converting an image to one of the same type performs a copy operation.
    """
    author = "Michael Droettboom, Karl MacMillan, and Christoph Dalitz"
    self_type = ImageType([ONEBIT, GREYSCALE, FLOAT, RGB, COMPLEX])
    return_type = ImageType([GREY16], "grey16")

    def __call__(self):
        if self.data.pixel_type == GREY16:
            return self.image_copy()
        return _image_conversion.to_grey16(self)
    __call__ = staticmethod(__call__)

class to_float(PluginFunction):
    """
    Converts the given image to a FLOAT image according to the following
    rules:

    - for ONEBIT images, 0 is mapped to 0.0 and everything else to 1.0
    - for GREYSCALE and GREY16 images, pixel values are copied unchanged
    - for RGB images, the luminance is used, which is defined in VIGRA as 0.3*R + 0.59*G + 0.11*B

    Converting an image to one of the same type performs a copy operation.
    """
    self_type = ImageType([ONEBIT, GREYSCALE, GREY16, RGB, COMPLEX])
    return_type = ImageType([FLOAT], "float")

    def __call__(self):
        if self.data.pixel_type == FLOAT:
            return self.image_copy()
        return _image_conversion.to_float(self)
    __call__ = staticmethod(__call__)

class to_complex(PluginFunction):
    """
    Converts the given image to a COMPLEX image.

    Note, converting an image to one of the same type performs a copy operation.
    """
    self_type = ImageType([ONEBIT, GREYSCALE, GREY16, RGB, FLOAT])
    return_type = ImageType([COMPLEX], "complex")

    def __call__(self):
        if self.data.pixel_type == COMPLEX:
            return self.image_copy()
        return _image_conversion.to_complex(self)
    __call__ = staticmethod(__call__)

class to_onebit(PluginFunction):
    """
    Converts the given image to a ONEBIT image. First the image is converted
    and then the otsu_threshold_ algorithm is applied.
    For other ways to convert to ONEBIT images, see the Binarization_ category.

    Converting an image to one of the same type performs a copy operation.

    .. _otsu_threshold: binarization.html#otsu-threshold
    .. _Binarization: binarization.html
    """
    pure_python = True
    self_type = ImageType([FLOAT, GREYSCALE, GREY16, RGB, COMPLEX])
    return_type = ImageType([ONEBIT], "onebit")
    def __call__(self, storage_format=DENSE):
        if self.data.pixel_type == ONEBIT:
            return self.image_copy()
        if self.data.pixel_type != GREYSCALE:
            self = _image_conversion.to_greyscale(self)
        return self.otsu_threshold(storage_format)
    __call__ = staticmethod(__call__)
    doc_examples = [(RGB,), (GREYSCALE,)]
to_onebit = to_onebit

class extract_real(PluginFunction):
    """
    Returns a Float image containing only the real values in the given
    complex image.
    """
    self_type = ImageType([COMPLEX])
    return_type = ImageType([FLOAT], "float")

class extract_imaginary(PluginFunction):
    """
    Returns a Float image containing only the imaginary values in the
    given complex image.
    """
    self_type = ImageType([COMPLEX])
    return_type = ImageType([FLOAT], "float")

class ImageConversionModule(PluginModule):
    category = "Conversion"
    cpp_headers=["image_conversion.hpp"]
    functions = [to_rgb, to_greyscale, to_grey16, to_float,
                 to_onebit, to_onebit, to_complex, extract_real,
                 extract_imaginary]
    author = "Michael Droettboom and Karl MacMillan"
    url = "http://gamera.sourceforge.net/"

module = ImageConversionModule()




#
# Copyright (C) 2012 Tobias Bolten
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#

import wx

from gamera.gui.gaoptimizer.ExpertSettingPanel import *

#-------------------------------------------------------------------------------
class ParallelizationPanel(ExpertSettingPanel):
#-------------------------------------------------------------------------------
    #---------------------------------------------------------------------------
    def __init__(self, parent, id):
    #---------------------------------------------------------------------------
        ExpertSettingPanel.__init__(self, parent, id)

        # sizer for widget setting
        sizer = wx.FlexGridSizer(0, 2, 5, 5)
        sizer.AddGrowableCol(1)
        self.SetSizer(sizer)

        # enable or disable the parallelization
        text = wx.StaticText(self, -1, "Status:")
        sizer.Add(text, 0, wx.ALIGN_BOTTOM | wx.LEFT | wx.RIGHT | wx.TOP, 10)
        self.parallelEnabled = wx.CheckBox(self, -1, "Enable Parallelization", \
            name = "parallelization")
        self.parallelEnabled.SetValue(True)
        self.parallelEnabled.SetToolTipString("Only available if compiled with OpenMP")
        sizer.Add(self.parallelEnabled, 0, wx.LEFT | wx.RIGHT | wx.TOP | wx.EXPAND, 10)

        # if enabled choose the number of used threads
        threadText = wx.StaticText(self, -1, "Number of Threads:")
        sizer.Add(threadText, 0, wx.ALIGN_BOTTOM | wx.LEFT | wx.RIGHT | wx.TOP, 10)

        try:
            # check available hardware
            import multiprocessing
            cpu_count = multiprocessing.cpu_count()
        except Exception:
            # when no multiprocessing module is available (python < 2.6)
            # assume 2 cores
            cpu_count = 2

        # by default use all instead of one core
        cores_to_use = cpu_count - 1
        if cores_to_use < 1:
            cores_to_use = 1

        self.threadNum = wx.Slider(self, -1, cores_to_use, 1, cpu_count, style= wx.SL_LABELS)

        sizer.Add(self.threadNum, 0, wx.LEFT | wx.RIGHT | wx.TOP | wx.EXPAND, 10)

        # save the logical dependencies
        self.genericWidgets.append(self.parallelEnabled)
        #self.AddChildToParent(self.parallelEnabled, threadText)
        self.AddChildToParent(self.parallelEnabled, self.threadNum)

        # bind the EVT_CHECKBOX to the CheckBox
        self.BindEvent(wx.EVT_CHECKBOX, self.OnCheckBox, [self.parallelEnabled])
